from copy import deepcopy
from fontTools.ttLib import newTable
from fontTools.ttLib.tables import otTables as ot
from fontTools.colorLib import builder
from fontTools.colorLib.geometry import round_start_circle_stable_containment, Circle
from fontTools.colorLib.builder import LayerListBuilder, buildCOLR
from fontTools.colorLib.table_builder import TableBuilder
from fontTools.colorLib.errors import ColorLibError
import pytest
from typing import List


def _build(cls, source):
    return LayerListBuilder().tableBuilder.build(cls, source)


def _buildPaint(source):
    return LayerListBuilder().buildPaint(source)


def test_buildCOLR_v0():
    color_layer_lists = {
        "a": [("a.color0", 0), ("a.color1", 1)],
        "b": [("b.color1", 1), ("b.color0", 0)],
    }

    colr = builder.buildCOLR(color_layer_lists)

    assert colr.tableTag == "COLR"
    assert colr.version == 0
    assert colr.ColorLayers["a"][0].name == "a.color0"
    assert colr.ColorLayers["a"][0].colorID == 0
    assert colr.ColorLayers["a"][1].name == "a.color1"
    assert colr.ColorLayers["a"][1].colorID == 1
    assert colr.ColorLayers["b"][0].name == "b.color1"
    assert colr.ColorLayers["b"][0].colorID == 1
    assert colr.ColorLayers["b"][1].name == "b.color0"
    assert colr.ColorLayers["b"][1].colorID == 0


def test_buildCOLR_v0_layer_as_list():
    # when COLRv0 layers are encoded as plist in UFO lib, both python tuples and
    # lists are encoded as plist array elements; but the latter are always decoded
    # as python lists, thus after roundtripping a plist tuples become lists.
    # Before FontTools 4.17.0 we were treating tuples and lists as equivalent;
    # with 4.17.0, a paint of type list is used to identify a PaintColrLayers.
    # This broke backward compatibility as ufo2ft is simply passing through the
    # color layers as read from the UFO lib plist, and as such the latter use lists
    # instead of tuples for COLRv0 layers (layerGlyph, paletteIndex) combo.
    # We restore backward compat by accepting either tuples or lists (of length 2
    # and only containing a str and an int) as individual top-level layers.
    # https://github.com/googlefonts/ufo2ft/issues/426
    color_layer_lists = {
        "a": [["a.color0", 0], ["a.color1", 1]],
        "b": [["b.color1", 1], ["b.color0", 0]],
    }

    colr = builder.buildCOLR(color_layer_lists)

    assert colr.tableTag == "COLR"
    assert colr.version == 0
    assert colr.ColorLayers["a"][0].name == "a.color0"
    assert colr.ColorLayers["a"][0].colorID == 0
    assert colr.ColorLayers["a"][1].name == "a.color1"
    assert colr.ColorLayers["a"][1].colorID == 1
    assert colr.ColorLayers["b"][0].name == "b.color1"
    assert colr.ColorLayers["b"][0].colorID == 1
    assert colr.ColorLayers["b"][1].name == "b.color0"
    assert colr.ColorLayers["b"][1].colorID == 0


def test_buildCPAL_v0():
    palettes = [
        [(0.68, 0.20, 0.32, 1.0), (0.45, 0.68, 0.21, 1.0)],
        [(0.68, 0.20, 0.32, 0.6), (0.45, 0.68, 0.21, 0.6)],
        [(0.68, 0.20, 0.32, 0.3), (0.45, 0.68, 0.21, 0.3)],
    ]

    cpal = builder.buildCPAL(palettes)

    assert cpal.tableTag == "CPAL"
    assert cpal.version == 0
    assert cpal.numPaletteEntries == 2

    assert len(cpal.palettes) == 3
    assert [tuple(c) for c in cpal.palettes[0]] == [
        (82, 51, 173, 255),
        (54, 173, 115, 255),
    ]
    assert [tuple(c) for c in cpal.palettes[1]] == [
        (82, 51, 173, 153),
        (54, 173, 115, 153),
    ]
    assert [tuple(c) for c in cpal.palettes[2]] == [
        (82, 51, 173, 76),
        (54, 173, 115, 76),
    ]


def test_buildCPAL_palettes_different_lengths():
    with pytest.raises(ColorLibError, match="have different lengths"):
        builder.buildCPAL([[(1, 1, 1, 1)], [(0, 0, 0, 1), (0.5, 0.5, 0.5, 1)]])


def test_buildPaletteLabels():
    name_table = newTable("name")
    name_table.names = []

    name_ids = builder.buildPaletteLabels(
        [None, "hi", {"en": "hello", "de": "hallo"}], name_table
    )

    assert name_ids == [0xFFFF, 256, 257]

    assert len(name_table.names) == 3
    assert str(name_table.names[0]) == "hi"
    assert name_table.names[0].nameID == 256

    assert str(name_table.names[1]) == "hallo"
    assert name_table.names[1].nameID == 257

    assert str(name_table.names[2]) == "hello"
    assert name_table.names[2].nameID == 257


def test_build_CPAL_v1_types_no_labels():
    palettes = [
        [(0.1, 0.2, 0.3, 1.0), (0.4, 0.5, 0.6, 1.0)],
        [(0.1, 0.2, 0.3, 0.6), (0.4, 0.5, 0.6, 0.6)],
        [(0.1, 0.2, 0.3, 0.3), (0.4, 0.5, 0.6, 0.3)],
    ]
    paletteTypes = [
        builder.ColorPaletteType.USABLE_WITH_LIGHT_BACKGROUND,
        builder.ColorPaletteType.USABLE_WITH_DARK_BACKGROUND,
        builder.ColorPaletteType.USABLE_WITH_LIGHT_BACKGROUND
        | builder.ColorPaletteType.USABLE_WITH_DARK_BACKGROUND,
    ]

    cpal = builder.buildCPAL(palettes, paletteTypes=paletteTypes)

    assert cpal.tableTag == "CPAL"
    assert cpal.version == 1
    assert cpal.numPaletteEntries == 2
    assert len(cpal.palettes) == 3

    assert cpal.paletteTypes == paletteTypes
    assert cpal.paletteLabels == [cpal.NO_NAME_ID] * len(palettes)
    assert cpal.paletteEntryLabels == [cpal.NO_NAME_ID] * cpal.numPaletteEntries


def test_build_CPAL_v1_labels():
    palettes = [
        [(0.1, 0.2, 0.3, 1.0), (0.4, 0.5, 0.6, 1.0)],
        [(0.1, 0.2, 0.3, 0.6), (0.4, 0.5, 0.6, 0.6)],
        [(0.1, 0.2, 0.3, 0.3), (0.4, 0.5, 0.6, 0.3)],
    ]
    paletteLabels = ["First", {"en": "Second", "it": "Seconda"}, None]
    paletteEntryLabels = ["Foo", "Bar"]

    with pytest.raises(TypeError, match="nameTable is required"):
        builder.buildCPAL(palettes, paletteLabels=paletteLabels)
    with pytest.raises(TypeError, match="nameTable is required"):
        builder.buildCPAL(palettes, paletteEntryLabels=paletteEntryLabels)

    name_table = newTable("name")
    name_table.names = []

    cpal = builder.buildCPAL(
        palettes,
        paletteLabels=paletteLabels,
        paletteEntryLabels=paletteEntryLabels,
        nameTable=name_table,
    )

    assert cpal.tableTag == "CPAL"
    assert cpal.version == 1
    assert cpal.numPaletteEntries == 2
    assert len(cpal.palettes) == 3

    assert cpal.paletteTypes == [cpal.DEFAULT_PALETTE_TYPE] * len(palettes)
    assert cpal.paletteLabels == [256, 257, cpal.NO_NAME_ID]
    assert cpal.paletteEntryLabels == [258, 259]

    assert name_table.getDebugName(256) == "First"
    assert name_table.getDebugName(257) == "Second"
    assert name_table.getDebugName(258) == "Foo"
    assert name_table.getDebugName(259) == "Bar"


def test_invalid_ColorPaletteType():
    with pytest.raises(ValueError, match="not a valid ColorPaletteType"):
        builder.ColorPaletteType(-1)
    with pytest.raises(ValueError, match="not a valid ColorPaletteType"):
        builder.ColorPaletteType(4)
    with pytest.raises(ValueError, match="not a valid ColorPaletteType"):
        builder.ColorPaletteType("abc")


def test_buildCPAL_v1_invalid_args_length():
    with pytest.raises(ColorLibError, match="Expected 2 paletteTypes, got 1"):
        builder.buildCPAL([[(0, 0, 0, 0)], [(1, 1, 1, 1)]], paletteTypes=[1])

    with pytest.raises(ColorLibError, match="Expected 2 paletteLabels, got 1"):
        builder.buildCPAL(
            [[(0, 0, 0, 0)], [(1, 1, 1, 1)]],
            paletteLabels=["foo"],
            nameTable=newTable("name"),
        )

    with pytest.raises(ColorLibError, match="Expected 1 paletteEntryLabels, got 0"):
        cpal = builder.buildCPAL(
            [[(0, 0, 0, 0)], [(1, 1, 1, 1)]],
            paletteEntryLabels=[],
            nameTable=newTable("name"),
        )


def test_buildCPAL_invalid_color():
    with pytest.raises(
        ColorLibError,
        match=r"In palette\[0\]\[1\]: expected \(R, G, B, A\) tuple, got \(1, 1, 1\)",
    ):
        builder.buildCPAL([[(1, 1, 1, 1), (1, 1, 1)]])

    with pytest.raises(
        ColorLibError,
        match=(
            r"palette\[1\]\[0\] has invalid out-of-range "
            r"\[0..1\] color: \(1, 1, -1, 2\)"
        ),
    ):
        builder.buildCPAL([[(0, 0, 0, 0)], [(1, 1, -1, 2)]])


def test_buildPaintSolid():
    p = _buildPaint((ot.PaintFormat.PaintSolid, 0))
    assert p.Format == ot.PaintFormat.PaintSolid
    assert p.PaletteIndex == 0
    assert p.Alpha == 1.0


def test_buildPaintSolid_Alpha():
    p = _buildPaint((ot.PaintFormat.PaintSolid, 1, 0.5))
    assert p.Format == ot.PaintFormat.PaintSolid
    assert p.PaletteIndex == 1
    assert p.Alpha == 0.5


def test_buildPaintVarSolid():
    p = _buildPaint((ot.PaintFormat.PaintVarSolid, 3, 0.5, 2))
    assert p.Format == ot.PaintFormat.PaintVarSolid
    assert p.PaletteIndex == 3
    assert p.Alpha == 0.5
    assert p.VarIndexBase == 2


def test_buildVarColorStop_DefaultAlpha():
    s = _build(ot.ColorStop, (0.1, 2))
    assert s.StopOffset == 0.1
    assert s.PaletteIndex == 2
    assert s.Alpha == builder._DEFAULT_ALPHA


def test_buildVarColorStop_DefaultAlpha():
    s = _build(ot.VarColorStop, (0.1, 2))
    assert s.StopOffset == 0.1
    assert s.PaletteIndex == 2
    assert s.Alpha == builder._DEFAULT_ALPHA


def test_buildColorStop():
    s = _build(ot.ColorStop, {"StopOffset": 0.2, "PaletteIndex": 3, "Alpha": 0.4})
    assert s.StopOffset == 0.2
    assert s.PaletteIndex == 3
    assert s.Alpha == 0.4


def test_buildColorStop_Variable():
    s = _build(
        ot.VarColorStop,
        {
            "StopOffset": 0.0,
            "PaletteIndex": 0,
            "Alpha": 0.3,
            "VarIndexBase": 1,
        },
    )
    assert s.StopOffset == 0.0
    assert s.PaletteIndex == 0
    assert s.Alpha == 0.3
    assert s.VarIndexBase == 1


def test_buildColorLine_StopList():
    stops = [(0.0, 0), (0.5, 1), (1.0, 2)]

    cline = _build(ot.ColorLine, {"ColorStop": stops})
    assert cline.Extend == builder.ExtendMode.PAD
    assert cline.StopCount == 3
    assert [(cs.StopOffset, cs.PaletteIndex) for cs in cline.ColorStop] == stops

    cline = _build(ot.ColorLine, {"Extend": "pad", "ColorStop": stops})
    assert cline.Extend == builder.ExtendMode.PAD

    cline = _build(
        ot.ColorLine, {"ColorStop": stops, "Extend": builder.ExtendMode.REPEAT}
    )
    assert cline.Extend == builder.ExtendMode.REPEAT

    cline = _build(
        ot.ColorLine, {"ColorStop": stops, "Extend": builder.ExtendMode.REFLECT}
    )
    assert cline.Extend == builder.ExtendMode.REFLECT

    cline = _build(
        ot.ColorLine, {"ColorStop": [_build(ot.ColorStop, s) for s in stops]}
    )
    assert [(cs.StopOffset, cs.PaletteIndex) for cs in cline.ColorStop] == stops


def test_buildVarColorLine_StopMap():
    stops = [
        {"StopOffset": 0.0, "PaletteIndex": 0, "Alpha": 0.5, "VarIndexBase": 1},
        {"StopOffset": 1.0, "PaletteIndex": 1, "Alpha": 0.3, "VarIndexBase": 3},
    ]
    cline = _build(ot.VarColorLine, {"ColorStop": stops})
    assert [
        {
            "StopOffset": cs.StopOffset,
            "PaletteIndex": cs.PaletteIndex,
            "Alpha": cs.Alpha,
            "VarIndexBase": cs.VarIndexBase,
        }
        for cs in cline.ColorStop
    ] == stops


def checkBuildAffine2x3(cls, variable=False):
    matrix = _build(cls, (1.5, 0, 0.5, 2.0, 1.0, -3.0))
    assert matrix.xx == 1.5
    assert matrix.yx == 0.0
    assert matrix.xy == 0.5
    assert matrix.yy == 2.0
    assert matrix.dx == 1.0
    assert matrix.dy == -3.0
    if variable:
        assert matrix.VarIndexBase == 0xFFFFFFFF


def test_buildAffine2x3():
    checkBuildAffine2x3(ot.Affine2x3)


def test_buildVarAffine2x3():
    checkBuildAffine2x3(ot.VarAffine2x3, variable=True)


def _sample_stops(variable):
    cls = ot.ColorStop if not variable else ot.VarColorStop
    stop_sources = [
        {"StopOffset": 0.0, "PaletteIndex": 0},
        {"StopOffset": 0.5, "PaletteIndex": 1},
        {"StopOffset": 1.0, "PaletteIndex": 2, "Alpha": 0.8},
    ]
    if variable:
        for i, src in enumerate(stop_sources, start=123):
            src["VarIndexBase"] = i
    return [_build(cls, src) for src in stop_sources]


def _is_var(fmt):
    return fmt.name.startswith("PaintVar")


def _is_around_center(fmt):
    return fmt.name.endswith("AroundCenter")


def _is_uniform_scale(fmt):
    return "ScaleUniform" in fmt.name


def checkBuildPaintLinearGradient(fmt):
    variable = _is_var(fmt)
    color_stops = _sample_stops(variable)

    x0, y0, x1, y1, x2, y2 = (1, 2, 3, 4, 5, 6)
    source = {
        "Format": fmt,
        "ColorLine": {"ColorStop": color_stops},
        "x0": x0,
        "y0": y0,
        "x1": x1,
        "y1": y1,
        "x2": x2,
        "y2": y2,
    }
    if variable:
        source["VarIndexBase"] = 7
    gradient = _buildPaint(source)
    assert gradient.ColorLine.Extend == builder.ExtendMode.PAD
    assert gradient.ColorLine.ColorStop == color_stops

    gradient = _buildPaint(gradient)
    assert (gradient.x0, gradient.y0) == (1, 2)
    assert (gradient.x1, gradient.y1) == (3, 4)
    assert (gradient.x2, gradient.y2) == (5, 6)
    if variable:
        assert gradient.VarIndexBase == 7


def test_buildPaintLinearGradient():
    assert not _is_var(ot.PaintFormat.PaintLinearGradient)
    checkBuildPaintLinearGradient(ot.PaintFormat.PaintLinearGradient)


def test_buildPaintVarLinearGradient():
    assert _is_var(ot.PaintFormat.PaintVarLinearGradient)
    checkBuildPaintLinearGradient(ot.PaintFormat.PaintVarLinearGradient)


def checkBuildPaintRadialGradient(fmt):
    variable = _is_var(fmt)
    color_stops = _sample_stops(variable)
    line_cls = ot.VarColorLine if variable else ot.ColorLine

    color_line = _build(
        line_cls, {"ColorStop": color_stops, "Extend": builder.ExtendMode.REPEAT}
    )
    c0 = (100, 200)
    c1 = (150, 250)
    r0 = 10
    r1 = 5
    varIndexBase = 0

    source = [fmt, color_line, *c0, r0, *c1, r1]
    if variable:
        source.append(varIndexBase)

    gradient = _build(ot.Paint, tuple(source))
    assert gradient.Format == fmt
    assert gradient.ColorLine == color_line
    assert (gradient.x0, gradient.y0) == c0
    assert (gradient.x1, gradient.y1) == c1
    assert gradient.r0 == r0
    assert gradient.r1 == r1
    if variable:
        assert gradient.VarIndexBase == varIndexBase

    source = {
        "Format": fmt,
        "ColorLine": {"ColorStop": color_stops},
        "x0": c0[0],
        "y0": c0[1],
        "x1": c1[0],
        "y1": c1[1],
        "r0": r0,
        "r1": r1,
    }
    if variable:
        source["VarIndexBase"] = varIndexBase
    gradient = _build(ot.Paint, source)
    assert gradient.ColorLine.Extend == builder.ExtendMode.PAD
    assert gradient.ColorLine.ColorStop == color_stops
    assert (gradient.x0, gradient.y0) == c0
    assert (gradient.x1, gradient.y1) == c1
    assert gradient.r0 == r0
    assert gradient.r1 == r1
    if variable:
        assert gradient.VarIndexBase == varIndexBase


def test_buildPaintRadialGradient():
    assert not _is_var(ot.PaintFormat.PaintRadialGradient)
    checkBuildPaintRadialGradient(ot.PaintFormat.PaintRadialGradient)


def test_buildPaintVarRadialGradient():
    assert _is_var(ot.PaintFormat.PaintVarRadialGradient)
    checkBuildPaintRadialGradient(ot.PaintFormat.PaintVarRadialGradient)


def checkPaintSweepGradient(fmt):
    variable = _is_var(fmt)
    source = {
        "Format": fmt,
        "ColorLine": {"ColorStop": _sample_stops(variable)},
        "centerX": 127,
        "centerY": 129,
        "startAngle": 15,
        "endAngle": 42,
    }
    if variable:
        source["VarIndexBase"] = 666
    paint = _buildPaint(source)

    assert paint.Format == fmt
    assert paint.centerX == 127
    assert paint.centerY == 129
    assert paint.startAngle == 15
    assert paint.endAngle == 42
    if variable:
        assert paint.VarIndexBase == 666


def test_buildPaintSweepGradient():
    assert not _is_var(ot.PaintFormat.PaintSweepGradient)
    checkPaintSweepGradient(ot.PaintFormat.PaintSweepGradient)


def test_buildPaintVarSweepGradient():
    assert _is_var(ot.PaintFormat.PaintVarSweepGradient)
    checkPaintSweepGradient(ot.PaintFormat.PaintVarSweepGradient)


def test_buildPaintGlyph_Solid():
    layer = _build(
        ot.Paint,
        (
            ot.PaintFormat.PaintGlyph,
            (
                ot.PaintFormat.PaintSolid,
                2,
            ),
            "a",
        ),
    )
    assert layer.Format == ot.PaintFormat.PaintGlyph
    assert layer.Glyph == "a"
    assert layer.Paint.Format == ot.PaintFormat.PaintSolid
    assert layer.Paint.PaletteIndex == 2

    layer = _build(
        ot.Paint,
        (
            ot.PaintFormat.PaintGlyph,
            (ot.PaintFormat.PaintSolid, 3, 0.9),
            "a",
        ),
    )
    assert layer.Paint.Format == ot.PaintFormat.PaintSolid
    assert layer.Paint.PaletteIndex == 3
    assert layer.Paint.Alpha == 0.9


def test_buildPaintGlyph_VarLinearGradient():
    layer = _build(
        ot.Paint,
        {
            "Format": ot.PaintFormat.PaintGlyph,
            "Glyph": "a",
            "Paint": {
                "Format": ot.PaintFormat.PaintVarLinearGradient,
                "ColorLine": {"ColorStop": [(0.0, 3), (1.0, 4)]},
                "x0": 100,
                "y0": 200,
                "x1": 150,
                "y1": 250,
            },
        },
    )

    assert layer.Format == ot.PaintFormat.PaintGlyph
    assert layer.Glyph == "a"
    assert layer.Paint.Format == ot.PaintFormat.PaintVarLinearGradient
    assert layer.Paint.ColorLine.ColorStop[0].StopOffset == 0.0
    assert layer.Paint.ColorLine.ColorStop[0].PaletteIndex == 3
    assert layer.Paint.ColorLine.ColorStop[1].StopOffset == 1.0
    assert layer.Paint.ColorLine.ColorStop[1].PaletteIndex == 4
    assert layer.Paint.x0 == 100
    assert layer.Paint.y0 == 200
    assert layer.Paint.x1 == 150
    assert layer.Paint.y1 == 250


def test_buildPaintGlyph_RadialGradient():
    layer = _build(
        ot.Paint,
        (
            int(ot.PaintFormat.PaintGlyph),
            (
                ot.PaintFormat.PaintRadialGradient,
                (
                    "pad",
                    [
                        (0.0, 5),
                        {"StopOffset": 0.5, "PaletteIndex": 6, "Alpha": 0.8},
                        (1.0, 7),
                    ],
                ),
                50,
                50,
                30,
                75,
                75,
                10,
            ),
            "a",
        ),
    )
    assert layer.Format == ot.PaintFormat.PaintGlyph
    assert layer.Paint.Format == ot.PaintFormat.PaintRadialGradient
    assert layer.Paint.ColorLine.ColorStop[0].StopOffset == 0.0
    assert layer.Paint.ColorLine.ColorStop[0].PaletteIndex == 5
    assert layer.Paint.ColorLine.ColorStop[1].StopOffset == 0.5
    assert layer.Paint.ColorLine.ColorStop[1].PaletteIndex == 6
    assert layer.Paint.ColorLine.ColorStop[1].Alpha == 0.8
    assert layer.Paint.ColorLine.ColorStop[2].StopOffset == 1.0
    assert layer.Paint.ColorLine.ColorStop[2].PaletteIndex == 7
    assert layer.Paint.x0 == 50
    assert layer.Paint.y0 == 50
    assert layer.Paint.r0 == 30
    assert layer.Paint.x1 == 75
    assert layer.Paint.y1 == 75
    assert layer.Paint.r1 == 10


def test_buildPaintGlyph_Dict_Solid():
    layer = _build(
        ot.Paint,
        (
            int(ot.PaintFormat.PaintGlyph),
            (int(ot.PaintFormat.PaintSolid), 1),
            "a",
        ),
    )
    assert layer.Format == ot.PaintFormat.PaintGlyph
    assert layer.Format == ot.PaintFormat.PaintGlyph
    assert layer.Glyph == "a"
    assert layer.Paint.Format == ot.PaintFormat.PaintSolid
    assert layer.Paint.PaletteIndex == 1


def test_buildPaintGlyph_Dict_VarLinearGradient():
    layer = _build(
        ot.Paint,
        {
            "Format": ot.PaintFormat.PaintGlyph,
            "Glyph": "a",
            "Paint": {
                "Format": int(ot.PaintFormat.PaintVarLinearGradient),
                "ColorLine": {"ColorStop": [(0.0, 0), (1.0, 1)]},
                "x0": 0,
                "y0": 0,
                "x1": 10,
                "y1": 10,
            },
        },
    )
    assert layer.Format == ot.PaintFormat.PaintGlyph
    assert layer.Glyph == "a"
    assert layer.Paint.Format == ot.PaintFormat.PaintVarLinearGradient
    assert layer.Paint.ColorLine.ColorStop[0].StopOffset == 0.0


def test_buildPaintGlyph_Dict_RadialGradient():
    layer = _buildPaint(
        {
            "Glyph": "a",
            "Paint": {
                "Format": int(ot.PaintFormat.PaintRadialGradient),
                "ColorLine": {"ColorStop": [(0.0, 0), (1.0, 1)]},
                "x0": 0,
                "y0": 0,
                "r0": 4,
                "x1": 10,
                "y1": 10,
                "r1": 0,
            },
            "Format": int(ot.PaintFormat.PaintGlyph),
        },
    )
    assert layer.Paint.Format == ot.PaintFormat.PaintRadialGradient
    assert layer.Paint.r0 == 4


def test_buildPaintColrGlyph():
    paint = _buildPaint((int(ot.PaintFormat.PaintColrGlyph), "a"))
    assert paint.Format == ot.PaintFormat.PaintColrGlyph
    assert paint.Glyph == "a"


def checkBuildPaintTransform(fmt):
    variable = _is_var(fmt)
    if variable:
        affine_cls = ot.VarAffine2x3
    else:
        affine_cls = ot.Affine2x3

    affine_src = [1, 2, 3, 4, 5, 6]
    if variable:
        affine_src.append(7)

    paint = _buildPaint(
        (
            int(fmt),
            (ot.PaintFormat.PaintGlyph, (ot.PaintFormat.PaintSolid, 0, 1.0), "a"),
            _build(affine_cls, tuple(affine_src)),
        ),
    )

    assert paint.Format == fmt
    assert paint.Paint.Format == ot.PaintFormat.PaintGlyph
    assert paint.Paint.Paint.Format == ot.PaintFormat.PaintSolid

    assert paint.Transform.xx == 1.0
    assert paint.Transform.yx == 2.0
    assert paint.Transform.xy == 3.0
    assert paint.Transform.yy == 4.0
    assert paint.Transform.dx == 5.0
    assert paint.Transform.dy == 6.0
    if variable:
        assert paint.Transform.VarIndexBase == 7

    affine_src = [1, 2, 3, 0.3333, 10, 10]
    if variable:
        affine_src.append(456)  # VarIndexBase
    paint = _build(
        ot.Paint,
        {
            "Format": fmt,
            "Transform": tuple(affine_src),
            "Paint": {
                "Format": int(ot.PaintFormat.PaintRadialGradient),
                "ColorLine": {"ColorStop": [(0.0, 0), (1.0, 1)]},
                "x0": 100,
                "y0": 101,
                "x1": 102,
                "y1": 103,
                "r0": 0,
                "r1": 50,
            },
        },
    )

    assert paint.Format == fmt
    assert paint.Transform.xx == 1.0
    assert paint.Transform.yx == 2.0
    assert paint.Transform.xy == 3.0
    assert paint.Transform.yy == 0.3333
    assert paint.Transform.dx == 10
    assert paint.Transform.dy == 10
    if variable:
        assert paint.Transform.VarIndexBase == 456
    assert paint.Paint.Format == ot.PaintFormat.PaintRadialGradient


def test_buildPaintTransform():
    assert not _is_var(ot.PaintFormat.PaintTransform)
    checkBuildPaintTransform(ot.PaintFormat.PaintTransform)


def test_buildPaintVarTransform():
    assert _is_var(ot.PaintFormat.PaintVarTransform)
    checkBuildPaintTransform(ot.PaintFormat.PaintVarTransform)


def test_buildPaintComposite():
    composite = _build(
        ot.Paint,
        {
            "Format": int(ot.PaintFormat.PaintComposite),
            "CompositeMode": "src_over",
            "SourcePaint": {
                "Format": ot.PaintFormat.PaintComposite,
                "CompositeMode": "src_over",
                "SourcePaint": {
                    "Format": int(ot.PaintFormat.PaintGlyph),
                    "Glyph": "c",
                    "Paint": (ot.PaintFormat.PaintSolid, 2),
                },
                "BackdropPaint": {
                    "Format": int(ot.PaintFormat.PaintGlyph),
                    "Glyph": "b",
                    "Paint": (ot.PaintFormat.PaintSolid, 1),
                },
            },
            "BackdropPaint": {
                "Format": ot.PaintFormat.PaintGlyph,
                "Glyph": "a",
                "Paint": {
                    "Format": ot.PaintFormat.PaintSolid,
                    "PaletteIndex": 0,
                    "Alpha": 0.5,
                },
            },
        },
    )

    assert composite.Format == ot.PaintFormat.PaintComposite
    assert composite.SourcePaint.Format == ot.PaintFormat.PaintComposite
    assert composite.SourcePaint.SourcePaint.Format == ot.PaintFormat.PaintGlyph
    assert composite.SourcePaint.SourcePaint.Glyph == "c"
    assert composite.SourcePaint.SourcePaint.Paint.Format == ot.PaintFormat.PaintSolid
    assert composite.SourcePaint.SourcePaint.Paint.PaletteIndex == 2
    assert composite.SourcePaint.CompositeMode == ot.CompositeMode.SRC_OVER
    assert composite.SourcePaint.BackdropPaint.Format == ot.PaintFormat.PaintGlyph
    assert composite.SourcePaint.BackdropPaint.Glyph == "b"
    assert composite.SourcePaint.BackdropPaint.Paint.Format == ot.PaintFormat.PaintSolid
    assert composite.SourcePaint.BackdropPaint.Paint.PaletteIndex == 1
    assert composite.CompositeMode == ot.CompositeMode.SRC_OVER
    assert composite.BackdropPaint.Format == ot.PaintFormat.PaintGlyph
    assert composite.BackdropPaint.Glyph == "a"
    assert composite.BackdropPaint.Paint.Format == ot.PaintFormat.PaintSolid
    assert composite.BackdropPaint.Paint.PaletteIndex == 0
    assert composite.BackdropPaint.Paint.Alpha == 0.5


def checkBuildPaintTranslate(fmt):
    variable = _is_var(fmt)

    source = {
        "Format": fmt,
        "Paint": (
            ot.PaintFormat.PaintGlyph,
            (ot.PaintFormat.PaintSolid, 0, 1.0),
            "a",
        ),
        "dx": 123,
        "dy": -345,
    }
    if variable:
        source["VarIndexBase"] = 678

    paint = _build(ot.Paint, source)

    assert paint.Format == fmt
    assert paint.Paint.Format == ot.PaintFormat.PaintGlyph
    assert paint.dx == 123
    assert paint.dy == -345
    if variable:
        assert paint.VarIndexBase == 678


def test_buildPaintTranslate():
    assert not _is_var(ot.PaintFormat.PaintTranslate)
    checkBuildPaintTranslate(ot.PaintFormat.PaintTranslate)


def test_buildPaintVarTranslate():
    assert _is_var(ot.PaintFormat.PaintVarTranslate)
    checkBuildPaintTranslate(ot.PaintFormat.PaintVarTranslate)


def checkBuildPaintScale(fmt):
    variable = _is_var(fmt)
    around_center = _is_around_center(fmt)
    uniform = _is_uniform_scale(fmt)

    source = {
        "Format": fmt,
        "Paint": (
            ot.PaintFormat.PaintGlyph,
            (ot.PaintFormat.PaintSolid, 0, 1.0),
            "a",
        ),
    }
    if uniform:
        source["scale"] = 1.5
    else:
        source["scaleX"] = 1.0
        source["scaleY"] = 2.0
    if around_center:
        source["centerX"] = 127
        source["centerY"] = 129
    if variable:
        source["VarIndexBase"] = 666

    paint = _build(ot.Paint, source)

    assert paint.Format == fmt
    assert paint.Paint.Format == ot.PaintFormat.PaintGlyph
    if uniform:
        assert paint.scale == 1.5
    else:
        assert paint.scaleX == 1.0
        assert paint.scaleY == 2.0
    if around_center:
        assert paint.centerX == 127
        assert paint.centerY == 129
    if variable:
        assert paint.VarIndexBase == 666


def test_buildPaintScale():
    assert not _is_var(ot.PaintFormat.PaintScale)
    assert not _is_uniform_scale(ot.PaintFormat.PaintScale)
    assert not _is_around_center(ot.PaintFormat.PaintScale)
    checkBuildPaintScale(ot.PaintFormat.PaintScale)


def test_buildPaintVarScale():
    assert _is_var(ot.PaintFormat.PaintVarScale)
    assert not _is_uniform_scale(ot.PaintFormat.PaintVarScale)
    assert not _is_around_center(ot.PaintFormat.PaintVarScale)
    checkBuildPaintScale(ot.PaintFormat.PaintVarScale)


def test_buildPaintScaleAroundCenter():
    assert not _is_var(ot.PaintFormat.PaintScaleAroundCenter)
    assert not _is_uniform_scale(ot.PaintFormat.PaintScaleAroundCenter)
    assert _is_around_center(ot.PaintFormat.PaintScaleAroundCenter)
    checkBuildPaintScale(ot.PaintFormat.PaintScaleAroundCenter)


def test_buildPaintVarScaleAroundCenter():
    assert _is_var(ot.PaintFormat.PaintVarScaleAroundCenter)
    assert not _is_uniform_scale(ot.PaintFormat.PaintScaleAroundCenter)
    assert _is_around_center(ot.PaintFormat.PaintVarScaleAroundCenter)
    checkBuildPaintScale(ot.PaintFormat.PaintVarScaleAroundCenter)


def test_buildPaintScaleUniform():
    assert not _is_var(ot.PaintFormat.PaintScaleUniform)
    assert _is_uniform_scale(ot.PaintFormat.PaintScaleUniform)
    assert not _is_around_center(ot.PaintFormat.PaintScaleUniform)
    checkBuildPaintScale(ot.PaintFormat.PaintScaleUniform)


def test_buildPaintVarScaleUniform():
    assert _is_var(ot.PaintFormat.PaintVarScaleUniform)
    assert _is_uniform_scale(ot.PaintFormat.PaintVarScaleUniform)
    assert not _is_around_center(ot.PaintFormat.PaintVarScaleUniform)
    checkBuildPaintScale(ot.PaintFormat.PaintVarScaleUniform)


def test_buildPaintScaleUniformAroundCenter():
    assert not _is_var(ot.PaintFormat.PaintScaleUniformAroundCenter)
    assert _is_uniform_scale(ot.PaintFormat.PaintScaleUniformAroundCenter)
    assert _is_around_center(ot.PaintFormat.PaintScaleUniformAroundCenter)
    checkBuildPaintScale(ot.PaintFormat.PaintScaleUniformAroundCenter)


def test_buildPaintVarScaleUniformAroundCenter():
    assert _is_var(ot.PaintFormat.PaintVarScaleUniformAroundCenter)
    assert _is_uniform_scale(ot.PaintFormat.PaintVarScaleUniformAroundCenter)
    assert _is_around_center(ot.PaintFormat.PaintVarScaleUniformAroundCenter)
    checkBuildPaintScale(ot.PaintFormat.PaintVarScaleUniformAroundCenter)


def checkBuildPaintRotate(fmt):
    variable = _is_var(fmt)
    around_center = _is_around_center(fmt)

    source = {
        "Format": fmt,
        "Paint": (
            ot.PaintFormat.PaintGlyph,
            (ot.PaintFormat.PaintSolid, 0, 1.0),
            "a",
        ),
        "angle": 15,
    }
    if around_center:
        source["centerX"] = 127
        source["centerY"] = 129

    paint = _build(ot.Paint, source)

    assert paint.Format == fmt
    assert paint.Paint.Format == ot.PaintFormat.PaintGlyph
    assert paint.angle == 15
    if around_center:
        assert paint.centerX == 127
        assert paint.centerY == 129
    if variable:
        assert paint.VarIndexBase == 0xFFFFFFFF


def test_buildPaintRotate():
    assert not _is_var(ot.PaintFormat.PaintRotate)
    assert not _is_around_center(ot.PaintFormat.PaintRotate)
    checkBuildPaintRotate(ot.PaintFormat.PaintRotate)


def test_buildPaintVarRotate():
    assert _is_var(ot.PaintFormat.PaintVarRotate)
    assert not _is_around_center(ot.PaintFormat.PaintVarRotate)
    checkBuildPaintRotate(ot.PaintFormat.PaintVarRotate)


def test_buildPaintRotateAroundCenter():
    assert not _is_var(ot.PaintFormat.PaintRotateAroundCenter)
    assert _is_around_center(ot.PaintFormat.PaintRotateAroundCenter)
    checkBuildPaintRotate(ot.PaintFormat.PaintRotateAroundCenter)


def test_buildPaintVarRotateAroundCenter():
    assert _is_var(ot.PaintFormat.PaintVarRotateAroundCenter)
    assert _is_around_center(ot.PaintFormat.PaintVarRotateAroundCenter)
    checkBuildPaintRotate(ot.PaintFormat.PaintVarRotateAroundCenter)


def checkBuildPaintSkew(fmt):
    variable = _is_var(fmt)
    around_center = _is_around_center(fmt)

    source = {
        "Format": fmt,
        "Paint": (
            ot.PaintFormat.PaintGlyph,
            (ot.PaintFormat.PaintSolid, 0, 1.0),
            "a",
        ),
        "xSkewAngle": 15,
        "ySkewAngle": 42,
    }
    if around_center:
        source["centerX"] = 127
        source["centerY"] = 129
    if variable:
        source["VarIndexBase"] = 0

    paint = _build(ot.Paint, source)

    assert paint.Format == fmt
    assert paint.Paint.Format == ot.PaintFormat.PaintGlyph
    assert paint.xSkewAngle == 15
    assert paint.ySkewAngle == 42
    if around_center:
        assert paint.centerX == 127
        assert paint.centerY == 129
    if variable:
        assert paint.VarIndexBase == 0


def test_buildPaintSkew():
    assert not _is_var(ot.PaintFormat.PaintSkew)
    assert not _is_around_center(ot.PaintFormat.PaintSkew)
    checkBuildPaintSkew(ot.PaintFormat.PaintSkew)


def test_buildPaintVarSkew():
    assert _is_var(ot.PaintFormat.PaintVarSkew)
    assert not _is_around_center(ot.PaintFormat.PaintVarSkew)
    checkBuildPaintSkew(ot.PaintFormat.PaintVarSkew)


def test_buildPaintSkewAroundCenter():
    assert not _is_var(ot.PaintFormat.PaintSkewAroundCenter)
    assert _is_around_center(ot.PaintFormat.PaintSkewAroundCenter)
    checkBuildPaintSkew(ot.PaintFormat.PaintSkewAroundCenter)


def test_buildPaintVarSkewAroundCenter():
    assert _is_var(ot.PaintFormat.PaintVarSkewAroundCenter)
    assert _is_around_center(ot.PaintFormat.PaintVarSkewAroundCenter)
    checkBuildPaintSkew(ot.PaintFormat.PaintVarSkewAroundCenter)


def test_buildColrV1():
    colorGlyphs = {
        "a": (
            ot.PaintFormat.PaintColrLayers,
            [
                (ot.PaintFormat.PaintGlyph, (ot.PaintFormat.PaintSolid, 0), "b"),
                (ot.PaintFormat.PaintGlyph, (ot.PaintFormat.PaintVarSolid, 1), "c"),
            ],
        ),
        "d": (
            ot.PaintFormat.PaintColrLayers,
            [
                (
                    ot.PaintFormat.PaintGlyph,
                    {
                        "Format": int(ot.PaintFormat.PaintSolid),
                        "PaletteIndex": 2,
                        "Alpha": 0.8,
                    },
                    "e",
                ),
                (
                    ot.PaintFormat.PaintGlyph,
                    {
                        "Format": int(ot.PaintFormat.PaintVarRadialGradient),
                        "ColorLine": {
                            "ColorStop": [(0.0, 3), (1.0, 4)],
                            "Extend": "reflect",
                        },
                        "x0": 0,
                        "y0": 0,
                        "x1": 0,
                        "y1": 0,
                        "r0": 10,
                        "r1": 0,
                    },
                    "f",
                ),
            ],
        ),
        "g": (
            ot.PaintFormat.PaintColrLayers,
            [(ot.PaintFormat.PaintGlyph, (ot.PaintFormat.PaintSolid, 5), "h")],
        ),
    }
    glyphMap = {
        ".notdef": 0,
        "a": 4,
        "b": 3,
        "c": 2,
        "d": 1,
        "e": 5,
        "f": 6,
        "g": 7,
        "h": 8,
    }

    # TODO(anthrotype) should we split into two tests? - seems two distinct validations
    layers, baseGlyphs = builder.buildColrV1(colorGlyphs, glyphMap)
    assert baseGlyphs.BaseGlyphCount == len(colorGlyphs)
    assert baseGlyphs.BaseGlyphPaintRecord[0].BaseGlyph == "d"
    assert baseGlyphs.BaseGlyphPaintRecord[1].BaseGlyph == "a"
    assert baseGlyphs.BaseGlyphPaintRecord[2].BaseGlyph == "g"

    layers, baseGlyphs = builder.buildColrV1(colorGlyphs)
    assert baseGlyphs.BaseGlyphCount == len(colorGlyphs)
    assert baseGlyphs.BaseGlyphPaintRecord[0].BaseGlyph == "a"
    assert baseGlyphs.BaseGlyphPaintRecord[1].BaseGlyph == "d"
    assert baseGlyphs.BaseGlyphPaintRecord[2].BaseGlyph == "g"


def test_buildColrV1_more_than_255_paints():
    num_paints = 364
    colorGlyphs = {
        "a": (
            ot.PaintFormat.PaintColrLayers,
            [
                {
                    "Format": int(ot.PaintFormat.PaintGlyph),
                    "Paint": (ot.PaintFormat.PaintSolid, 0),
                    "Glyph": name,
                }
                for name in (f"glyph{i}" for i in range(num_paints))
            ],
        ),
    }
    layers, baseGlyphs = builder.buildColrV1(colorGlyphs)
    paints = layers.Paint

    assert len(paints) == num_paints + 1

    assert all(paints[i].Format == ot.PaintFormat.PaintGlyph for i in range(255))

    assert paints[255].Format == ot.PaintFormat.PaintColrLayers
    assert paints[255].FirstLayerIndex == 0
    assert paints[255].NumLayers == 255

    assert all(
        paints[i].Format == ot.PaintFormat.PaintGlyph
        for i in range(256, num_paints + 1)
    )

    assert baseGlyphs.BaseGlyphCount == len(colorGlyphs)
    assert baseGlyphs.BaseGlyphPaintRecord[0].BaseGlyph == "a"
    assert (
        baseGlyphs.BaseGlyphPaintRecord[0].Paint.Format
        == ot.PaintFormat.PaintColrLayers
    )
    assert baseGlyphs.BaseGlyphPaintRecord[0].Paint.FirstLayerIndex == 255
    assert baseGlyphs.BaseGlyphPaintRecord[0].Paint.NumLayers == num_paints + 1 - 255


def test_split_color_glyphs_by_version():
    layerBuilder = LayerListBuilder()
    colorGlyphs = {
        "a": [
            ("b", 0),
            ("c", 1),
            ("d", 2),
            ("e", 3),
        ]
    }

    colorGlyphsV0, colorGlyphsV1 = builder._split_color_glyphs_by_version(colorGlyphs)

    assert colorGlyphsV0 == {"a": [("b", 0), ("c", 1), ("d", 2), ("e", 3)]}
    assert not colorGlyphsV1

    colorGlyphs = {"a": (ot.PaintFormat.PaintGlyph, 0, "b")}

    colorGlyphsV0, colorGlyphsV1 = builder._split_color_glyphs_by_version(colorGlyphs)

    assert not colorGlyphsV0
    assert colorGlyphsV1 == colorGlyphs

    colorGlyphs = {
        "a": [("b", 0)],
        "c": [
            ("d", 1),
            (
                "e",
                {
                    "format": 3,
                    "colorLine": {"stops": [(0.0, 2), (1.0, 3)]},
                    "p0": (0, 0),
                    "p1": (10, 10),
                },
            ),
        ],
    }

    colorGlyphsV0, colorGlyphsV1 = builder._split_color_glyphs_by_version(colorGlyphs)

    assert colorGlyphsV0 == {"a": [("b", 0)]}
    assert "a" not in colorGlyphsV1
    assert "c" in colorGlyphsV1
    assert len(colorGlyphsV1["c"]) == 2


def assertIsColrV1(colr):
    assert colr.version == 1
    assert not hasattr(colr, "ColorLayers")
    assert hasattr(colr, "table")
    assert isinstance(colr.table, ot.COLR)


def assertNoV0Content(colr):
    assert colr.table.BaseGlyphRecordCount == 0
    assert colr.table.BaseGlyphRecordArray is None
    assert colr.table.LayerRecordCount == 0
    assert colr.table.LayerRecordArray is None


def test_build_layerv1list_empty():
    # Nobody uses PaintColrLayers, no layerlist
    colr = builder.buildCOLR(
        {
            # BaseGlyph, tuple form
            "a": (
                int(ot.PaintFormat.PaintGlyph),
                (int(ot.PaintFormat.PaintSolid), 2, 0.8),
                "b",
            ),
            # BaseGlyph, map form
            "b": {
                "Format": int(ot.PaintFormat.PaintGlyph),
                "Paint": {
                    "Format": int(ot.PaintFormat.PaintLinearGradient),
                    "ColorLine": {
                        "ColorStop": [(0.0, 2), (1.0, 3)],
                        "Extend": "reflect",
                    },
                    "x0": 1,
                    "y0": 2,
                    "x1": 3,
                    "y1": 4,
                    "x2": 2,
                    "y2": 2,
                },
                "Glyph": "bb",
            },
        },
        version=1,
    )

    assertIsColrV1(colr)
    assertNoV0Content(colr)

    # 2 v1 glyphs, none in LayerList
    assert colr.table.BaseGlyphList.BaseGlyphCount == 2
    assert len(colr.table.BaseGlyphList.BaseGlyphPaintRecord) == 2
    assert colr.table.LayerList is None


def _paint_names(paints) -> List[str]:
    # prints a predictable string from a paint list to enable
    # semi-readable assertions on a LayerList order.
    result = []
    for paint in paints:
        if paint.Format == int(ot.PaintFormat.PaintGlyph):
            result.append(paint.Glyph)
        elif paint.Format == int(ot.PaintFormat.PaintColrLayers):
            result.append(
                f"Layers[{paint.FirstLayerIndex}:{paint.FirstLayerIndex+paint.NumLayers}]"
            )
    return result


def test_build_layerv1list_simple():
    # Two colr glyphs, each with two layers the first of which is common
    # All layers use the same solid paint
    solid_paint = {
        "Format": int(ot.PaintFormat.PaintSolid),
        "PaletteIndex": 2,
        "Alpha": 0.8,
    }
    backdrop = {
        "Format": int(ot.PaintFormat.PaintGlyph),
        "Paint": solid_paint,
        "Glyph": "back",
    }
    a_foreground = {
        "Format": int(ot.PaintFormat.PaintGlyph),
        "Paint": solid_paint,
        "Glyph": "a_fore",
    }
    b_foreground = {
        "Format": int(ot.PaintFormat.PaintGlyph),
        "Paint": solid_paint,
        "Glyph": "b_fore",
    }

    # list => PaintColrLayers, contents should land in LayerList
    colr = builder.buildCOLR(
        {
            "a": (
                ot.PaintFormat.PaintColrLayers,
                [
                    backdrop,
                    a_foreground,
                ],
            ),
            "b": {
                "Format": ot.PaintFormat.PaintColrLayers,
                "Layers": [
                    backdrop,
                    b_foreground,
                ],
            },
        },
        version=1,
    )

    assertIsColrV1(colr)
    assertNoV0Content(colr)

    # 2 v1 glyphs, 4 paints in LayerList
    # A single shared backdrop isn't worth accessing by slice
    assert colr.table.BaseGlyphList.BaseGlyphCount == 2
    assert len(colr.table.BaseGlyphList.BaseGlyphPaintRecord) == 2
    assert colr.table.LayerList.LayerCount == 4
    assert _paint_names(colr.table.LayerList.Paint) == [
        "back",
        "a_fore",
        "back",
        "b_fore",
    ]


def test_build_layerv1list_with_sharing():
    # Three colr glyphs, each with two layers in common
    solid_paint = {
        "Format": int(ot.PaintFormat.PaintSolid),
        "PaletteIndex": 2,
        "Alpha": 0.8,
    }
    backdrop = [
        {
            "Format": int(ot.PaintFormat.PaintGlyph),
            "Paint": solid_paint,
            "Glyph": "back1",
        },
        {
            "Format": ot.PaintFormat.PaintGlyph,
            "Paint": solid_paint,
            "Glyph": "back2",
        },
    ]
    a_foreground = {
        "Format": ot.PaintFormat.PaintGlyph,
        "Paint": solid_paint,
        "Glyph": "a_fore",
    }
    b_background = {
        "Format": ot.PaintFormat.PaintGlyph,
        "Paint": solid_paint,
        "Glyph": "b_back",
    }
    b_foreground = {
        "Format": ot.PaintFormat.PaintGlyph,
        "Paint": solid_paint,
        "Glyph": "b_fore",
    }
    c_background = {
        "Format": ot.PaintFormat.PaintGlyph,
        "Paint": solid_paint,
        "Glyph": "c_back",
    }

    # list => PaintColrLayers, which means contents should be in LayerList
    colr = builder.buildCOLR(
        {
            "a": (ot.PaintFormat.PaintColrLayers, backdrop + [a_foreground]),
            "b": (
                ot.PaintFormat.PaintColrLayers,
                [b_background] + backdrop + [b_foreground],
            ),
            "c": (ot.PaintFormat.PaintColrLayers, [c_background] + backdrop),
        },
        version=1,
    )

    assertIsColrV1(colr)
    assertNoV0Content(colr)

    # 2 v1 glyphs, 4 paints in LayerList
    # A single shared backdrop isn't worth accessing by slice
    baseGlyphs = colr.table.BaseGlyphList.BaseGlyphPaintRecord
    assert colr.table.BaseGlyphList.BaseGlyphCount == 3
    assert len(baseGlyphs) == 3
    assert _paint_names([b.Paint for b in baseGlyphs]) == [
        "Layers[0:3]",
        "Layers[3:6]",
        "Layers[6:8]",
    ]
    assert _paint_names(colr.table.LayerList.Paint) == [
        "back1",
        "back2",
        "a_fore",
        "b_back",
        "Layers[0:2]",
        "b_fore",
        "c_back",
        "Layers[0:2]",
    ]
    assert colr.table.LayerList.LayerCount == 8


def test_build_layerv1list_with_overlaps():
    paints = [
        {
            "Format": ot.PaintFormat.PaintGlyph,
            "Paint": {
                "Format": ot.PaintFormat.PaintSolid,
                "PaletteIndex": 2,
                "Alpha": 0.8,
            },
            "Glyph": c,
        }
        for c in "abcdefghi"
    ]

    # list => PaintColrLayers, which means contents should be in LayerList
    colr = builder.buildCOLR(
        {
            "a": (ot.PaintFormat.PaintColrLayers, paints[0:4]),
            "b": (ot.PaintFormat.PaintColrLayers, paints[0:6]),
            "c": (ot.PaintFormat.PaintColrLayers, paints[2:8]),
        },
        version=1,
    )

    assertIsColrV1(colr)
    assertNoV0Content(colr)

    baseGlyphs = colr.table.BaseGlyphList.BaseGlyphPaintRecord
    # assert colr.table.BaseGlyphList.BaseGlyphCount == 2

    assert _paint_names(colr.table.LayerList.Paint) == [
        "a",
        "b",
        "c",
        "d",
        "Layers[0:4]",
        "e",
        "f",
        "Layers[2:4]",
        "Layers[5:7]",
        "g",
        "h",
    ]
    assert _paint_names([b.Paint for b in baseGlyphs]) == [
        "Layers[0:4]",
        "Layers[4:7]",
        "Layers[7:11]",
    ]
    assert colr.table.LayerList.LayerCount == 11


def test_explicit_version_1():
    colr = builder.buildCOLR(
        {
            "a": (
                ot.PaintFormat.PaintColrLayers,
                [
                    (ot.PaintFormat.PaintGlyph, (ot.PaintFormat.PaintSolid, 0), "b"),
                    (ot.PaintFormat.PaintGlyph, (ot.PaintFormat.PaintSolid, 1), "c"),
                ],
            )
        },
        version=1,
    )
    assert colr.version == 1
    assert not hasattr(colr, "ColorLayers")
    assert hasattr(colr, "table")
    assert isinstance(colr.table, ot.COLR)
    assert colr.table.VarStore is None


class BuildCOLRTest(object):
    def test_automatic_version_all_solid_color_glyphs(self):
        colr = builder.buildCOLR({"a": [("b", 0), ("c", 1)]})
        assert colr.version == 0
        assert hasattr(colr, "ColorLayers")
        assert colr.ColorLayers["a"][0].name == "b"
        assert colr.ColorLayers["a"][1].name == "c"

    def test_automatic_version_no_solid_color_glyphs(self):
        colr = builder.buildCOLR(
            {
                "a": (
                    ot.PaintFormat.PaintColrLayers,
                    [
                        (
                            ot.PaintFormat.PaintGlyph,
                            {
                                "Format": int(ot.PaintFormat.PaintRadialGradient),
                                "ColorLine": {
                                    "ColorStop": [(0.0, 0), (1.0, 1)],
                                    "Extend": "repeat",
                                },
                                "x0": 1,
                                "y0": 0,
                                "x1": 10,
                                "y1": 0,
                                "r0": 4,
                                "r1": 2,
                            },
                            "b",
                        ),
                        (
                            ot.PaintFormat.PaintGlyph,
                            {
                                "Format": ot.PaintFormat.PaintSolid,
                                "PaletteIndex": 2,
                                "Alpha": 0.8,
                            },
                            "c",
                        ),
                    ],
                ),
                "d": (
                    ot.PaintFormat.PaintColrLayers,
                    [
                        {
                            "Format": ot.PaintFormat.PaintGlyph,
                            "Glyph": "e",
                            "Paint": {
                                "Format": ot.PaintFormat.PaintLinearGradient,
                                "ColorLine": {
                                    "ColorStop": [(0.0, 2), (1.0, 3)],
                                    "Extend": "reflect",
                                },
                                "x0": 1,
                                "y0": 2,
                                "x1": 3,
                                "y1": 4,
                                "x2": 2,
                                "y2": 2,
                            },
                        }
                    ],
                ),
            }
        )
        assertIsColrV1(colr)
        assert colr.table.BaseGlyphRecordCount == 0
        assert colr.table.BaseGlyphRecordArray is None
        assert colr.table.LayerRecordCount == 0
        assert colr.table.LayerRecordArray is None

    def test_automatic_version_mixed_solid_and_gradient_glyphs(self):
        colr = builder.buildCOLR(
            {
                "a": [("b", 0), ("c", 1)],
                "d": (
                    ot.PaintFormat.PaintColrLayers,
                    [
                        (
                            ot.PaintFormat.PaintGlyph,
                            {
                                "Format": ot.PaintFormat.PaintLinearGradient,
                                "ColorLine": {"ColorStop": [(0.0, 2), (1.0, 3)]},
                                "x0": 1,
                                "y0": 2,
                                "x1": 3,
                                "y1": 4,
                                "x2": 2,
                                "y2": 2,
                            },
                            "e",
                        ),
                        (
                            ot.PaintFormat.PaintGlyph,
                            (ot.PaintFormat.PaintSolid, 2, 0.8),
                            "f",
                        ),
                    ],
                ),
            }
        )
        assertIsColrV1(colr)
        assert colr.table.VarStore is None

        assert colr.table.BaseGlyphRecordCount == 1
        assert isinstance(colr.table.BaseGlyphRecordArray, ot.BaseGlyphRecordArray)
        assert colr.table.LayerRecordCount == 2
        assert isinstance(colr.table.LayerRecordArray, ot.LayerRecordArray)

        assert isinstance(colr.table.BaseGlyphList, ot.BaseGlyphList)
        assert colr.table.BaseGlyphList.BaseGlyphCount == 1
        assert isinstance(
            colr.table.BaseGlyphList.BaseGlyphPaintRecord[0], ot.BaseGlyphPaintRecord
        )
        assert colr.table.BaseGlyphList.BaseGlyphPaintRecord[0].BaseGlyph == "d"
        assert isinstance(colr.table.LayerList, ot.LayerList)
        assert colr.table.LayerList.Paint[0].Glyph == "e"

    def test_explicit_version_0(self):
        colr = builder.buildCOLR({"a": [("b", 0), ("c", 1)]}, version=0)
        assert colr.version == 0
        assert hasattr(colr, "ColorLayers")

    def test_explicit_version_1(self):
        colr = builder.buildCOLR(
            {
                "a": (
                    ot.PaintFormat.PaintColrLayers,
                    [
                        (
                            ot.PaintFormat.PaintGlyph,
                            (ot.PaintFormat.PaintSolid, 0),
                            "b",
                        ),
                        (
                            ot.PaintFormat.PaintGlyph,
                            (ot.PaintFormat.PaintSolid, 1),
                            "c",
                        ),
                    ],
                )
            },
            version=1,
        )
        assert colr.version == 1
        assert not hasattr(colr, "ColorLayers")
        assert hasattr(colr, "table")
        assert isinstance(colr.table, ot.COLR)
        assert colr.table.VarStore is None

    def test_paint_one_colr_layers(self):
        # A set of one layers should flip to just that layer
        colr = builder.buildCOLR(
            {
                "a": (
                    ot.PaintFormat.PaintColrLayers,
                    [
                        (
                            ot.PaintFormat.PaintGlyph,
                            (ot.PaintFormat.PaintSolid, 0),
                            "b",
                        ),
                    ],
                )
            },
        )

        assert colr.table.LayerList is None, "PaintColrLayers should be gone"
        assert colr.table.BaseGlyphList.BaseGlyphCount == 1
        paint = colr.table.BaseGlyphList.BaseGlyphPaintRecord[0].Paint
        assert paint.Format == ot.PaintFormat.PaintGlyph
        assert paint.Paint.Format == ot.PaintFormat.PaintSolid

    def test_build_clip_list(self):
        colr = builder.buildCOLR(
            {
                "a": (
                    ot.PaintFormat.PaintGlyph,
                    (ot.PaintFormat.PaintSolid, 0),
                    "b",
                ),
                "c": (
                    ot.PaintFormat.PaintGlyph,
                    (ot.PaintFormat.PaintSolid, 1),
                    "d",
                ),
            },
            clipBoxes={
                "a": (0, 0, 1000, 1000, 0),  # optional 5th: varIndexBase
                "c": (-100.8, -200.4, 1100.1, 1200.5),  # floats get rounded
                "e": (0, 0, 10, 10),  # 'e' does _not_ get ignored despite being missing
            },
        )

        assert colr.table.ClipList.Format == 1
        clipBoxes = colr.table.ClipList.clips
        assert [
            (baseGlyph, clipBox.as_tuple()) for baseGlyph, clipBox in clipBoxes.items()
        ] == [
            ("a", (0, 0, 1000, 1000, 0)),
            ("c", (-101, -201, 1101, 1201)),
            ("e", (0, 0, 10, 10)),
        ]
        assert clipBoxes["a"].Format == 2
        assert clipBoxes["c"].Format == 1
        assert clipBoxes["e"].Format == 1

    def test_duplicate_base_glyphs(self):
        # If > 1 base glyphs refer to equivalent list of layers we expect them to share
        # the same PaintColrLayers.
        layers = {
            "Format": ot.PaintFormat.PaintColrLayers,
            "Layers": [
                (ot.PaintFormat.PaintGlyph, (ot.PaintFormat.PaintSolid, 0), "d"),
                (ot.PaintFormat.PaintGlyph, (ot.PaintFormat.PaintSolid, 1), "e"),
            ],
        }
        # I copy the layers to ensure equality is by content, not by identity
        colr = builder.buildCOLR(
            {"a": layers, "b": deepcopy(layers), "c": deepcopy(layers)}
        ).table

        baseGlyphs = colr.BaseGlyphList.BaseGlyphPaintRecord
        assert len(baseGlyphs) == 3

        assert baseGlyphs[0].BaseGlyph == "a"
        assert baseGlyphs[1].BaseGlyph == "b"
        assert baseGlyphs[2].BaseGlyph == "c"

        expected = {"Format": 1, "FirstLayerIndex": 0, "NumLayers": 2}
        assert baseGlyphs[0].Paint.__dict__ == expected
        assert baseGlyphs[1].Paint.__dict__ == expected
        assert baseGlyphs[2].Paint.__dict__ == expected


class TrickyRadialGradientTest:
    @staticmethod
    def circle_inside_circle(c0, r0, c1, r1, rounded=False):
        if rounded:
            return Circle(c0, r0).round().inside(Circle(c1, r1).round())
        else:
            return Circle(c0, r0).inside(Circle(c1, r1))

    def round_start_circle(self, c0, r0, c1, r1, inside=True):
        assert self.circle_inside_circle(c0, r0, c1, r1) is inside
        assert self.circle_inside_circle(c0, r0, c1, r1, rounded=True) is not inside
        r = round_start_circle_stable_containment(c0, r0, c1, r1)
        assert (
            self.circle_inside_circle(r.centre, r.radius, c1, r1, rounded=True)
            is inside
        )
        return r.centre, r.radius

    def test_noto_emoji_mosquito_u1f99f(self):
        # https://github.com/googlefonts/picosvg/issues/158
        c0 = (385.23508, 70.56727999999998)
        r0 = 0
        c1 = (642.99108, 104.70327999999995)
        r1 = 260.0072
        assert self.round_start_circle(c0, r0, c1, r1, inside=True) == ((386, 71), 0)

    def test_noto_emoji_horns_sign_u1f918_1f3fc(self):
        # This radial gradient is taken from noto-emoji's 'SIGNS OF THE HORNS'
        # (1f918_1f3fc). We check that c0 is inside c1 both before and after rounding.
        c0 = (-437.6789059060543, -2116.9237094478003)
        r0 = 0.0
        c1 = (-488.7330118252256, -1876.5036857045086)
        r1 = 245.77147821915673
        assert self.circle_inside_circle(c0, r0, c1, r1)
        assert self.circle_inside_circle(c0, r0, c1, r1, rounded=True)

    @pytest.mark.parametrize(
        "c0, r0, c1, r1, inside, expected",
        [
            # inside before round, outside after round
            ((1.4, 0), 0, (2.6, 0), 1.3, True, ((2, 0), 0)),
            ((1, 0), 0.6, (2.8, 0), 2.45, True, ((2, 0), 1)),
            ((6.49, 6.49), 0, (0.49, 0.49), 8.49, True, ((5, 5), 0)),
            # outside before round, inside after round
            ((0, 0), 0, (2, 0), 1.5, False, ((-1, 0), 0)),
            ((0, -0.5), 0, (0, -2.5), 1.5, False, ((0, 1), 0)),
            # the following ones require two nudges to round correctly
            ((0.5, 0), 0, (9.4, 0), 8.8, False, ((-1, 0), 0)),
            ((1.5, 1.5), 0, (0.49, 0.49), 1.49, True, ((0, 0), 0)),
            # limit case when circle almost exactly overlap
            ((0.5000001, 0), 0.5000001, (0.499999, 0), 0.4999999, True, ((0, 0), 0)),
            # concentrical circles, r0 > r1
            ((0, 0), 1.49, (0, 0), 1, False, ((0, 0), 2)),
        ],
    )
    def test_nudge_start_circle_position(self, c0, r0, c1, r1, inside, expected):
        assert self.round_start_circle(c0, r0, c1, r1, inside) == expected


def test_buildCOLR_missing_base_glyph_in_glyphMap_raises_clear_error():
    glyphMap = {"A": 0}  # pretend the font only contains glyph A

    # base glyph B does not exist in glyphMap
    colorGlyphs = {"B": [("A", 0)]}

    with pytest.raises(ColorLibError, match="base glyph\\(s\\) not found in glyphMap"):
        buildCOLR(colorGlyphs, version=0, glyphMap=glyphMap)



"""ttLib.macUtils.py -- Various Mac-specific stuff."""

from io import BytesIO
from fontTools.misc.macRes import ResourceReader, ResourceError


def getSFNTResIndices(path):
    """Determine whether a file has a 'sfnt' resource fork or not."""
    try:
        reader = ResourceReader(path)
        indices = reader.getIndices("sfnt")
        reader.close()
        return indices
    except ResourceError:
        return []


def openTTFonts(path):
    """Given a pathname, return a list of TTFont objects. In the case
    of a flat TTF/OTF file, the list will contain just one font object;
    but in the case of a Mac font suitcase it will contain as many
    font objects as there are sfnt resources in the file.
    """
    from fontTools import ttLib

    fonts = []
    sfnts = getSFNTResIndices(path)
    if not sfnts:
        fonts.append(ttLib.TTFont(path))
    else:
        for index in sfnts:
            fonts.append(ttLib.TTFont(path, index))
        if not fonts:
            raise ttLib.TTLibError("no fonts found in file '%s'" % path)
    return fonts


class SFNTResourceReader(BytesIO):
    """Simple read-only file wrapper for 'sfnt' resources."""

    def __init__(self, path, res_name_or_index):
        from fontTools import ttLib

        reader = ResourceReader(path)
        if isinstance(res_name_or_index, str):
            rsrc = reader.getNamedResource("sfnt", res_name_or_index)
        else:
            rsrc = reader.getIndResource("sfnt", res_name_or_index)
        if rsrc is None:
            raise ttLib.TTLibError("sfnt resource not found: %s" % res_name_or_index)
        reader.close()
        self.rsrc = rsrc
        super(SFNTResourceReader, self).__init__(rsrc.data)
        self.name = path



#!/usr/bin/env python3

import argparse
import logging
import os
import sys

from fontTools.ttLib import TTFont


def ProcessTable(table):
    found = set()

    for rec in table.ScriptList.ScriptRecord:
        if rec.ScriptTag == "DFLT" and rec.Script.LangSysCount != 0:
            tags = [r.LangSysTag for r in rec.Script.LangSysRecord]
            logging.info(
                "Removing %d extraneous LangSys records: %s",
                rec.Script.LangSysCount,
                " ".join(tags),
            )
            rec.Script.LangSysRecord = []
            rec.Script.LangSysCount = 0
            found.update(tags)

    if not found:
        logging.info("All fine")
        return False
    else:
        for rec in table.ScriptList.ScriptRecord:
            tags = set([r.LangSysTag for r in rec.Script.LangSysRecord])
            found -= tags

        if found:
            logging.warning(
                "Records are missing from non-DFLT scripts: %s", " ".join(found)
            )
        return True


def ProcessFont(font):
    found = False
    for tag in ("GSUB", "GPOS"):
        if tag in font:
            logging.info("Processing %s table", tag)
            if ProcessTable(font[tag].table):
                found = True
            else:
                # Unmark the table as loaded so that it is read from disk when
                # writing the font, to avoid any unnecessary changes caused by
                # decompiling then recompiling again.
                del font.tables[tag]

    return found


def ProcessFiles(filenames):
    for filename in filenames:
        logging.info("Processing %s", filename)
        font = TTFont(filename)
        name, ext = os.path.splitext(filename)
        fixedname = name + ".fixed" + ext
        if ProcessFont(font):
            logging.info("Saving fixed font to %s\n", fixedname)
            font.save(fixedname)
        else:
            logging.info("Font file is fine, nothing to fix\n")


def main():
    parser = argparse.ArgumentParser(description="Fix LangSys records for DFLT script")
    parser.add_argument(
        "files", metavar="FILE", type=str, nargs="+", help="input font to process"
    )
    parser.add_argument(
        "-s", "--silent", action="store_true", help="suppress normal messages"
    )

    args = parser.parse_args()

    logformat = "%(levelname)s: %(message)s"
    if args.silent:
        logging.basicConfig(format=logformat, level=logging.DEBUG)
    else:
        logging.basicConfig(format=logformat, level=logging.INFO)

    ProcessFiles(args.files)


if __name__ == "__main__":
    sys.exit(main())



from . import asciiTable


class table_T_T_F_A_(asciiTable.asciiTable):
    """ttfautohint parameters table

    The ``TTFA`` table is used by the free-software `ttfautohint` program
    to record the parameters that `ttfautohint` was called with when it
    was used to auto-hint the font.

    See also http://freetype.org/ttfautohint/doc/ttfautohint.html#miscellaneous-1
    """

    pass



from fontTools.ttLib.ttGlyphSet import LerpGlyphSet
from fontTools.pens.basePen import AbstractPen, BasePen, DecomposingPen
from fontTools.pens.pointPen import AbstractPointPen, SegmentToPointPen
from fontTools.pens.recordingPen import RecordingPen, DecomposingRecordingPen
from fontTools.misc.transform import Transform
from collections import defaultdict, deque
from math import sqrt, copysign, atan2, pi
import itertools

import logging

log = logging.getLogger("fontTools.varLib.interpolatable")


class InterpolatableProblem:
    NOTHING = "nothing"
    MISSING = "missing"
    OPEN_PATH = "open_path"
    PATH_COUNT = "path_count"
    NODE_COUNT = "node_count"
    NODE_INCOMPATIBILITY = "node_incompatibility"
    CONTOUR_ORDER = "contour_order"
    WRONG_START_POINT = "wrong_start_point"
    KINK = "kink"
    UNDERWEIGHT = "underweight"
    OVERWEIGHT = "overweight"

    severity = {
        MISSING: 1,
        OPEN_PATH: 2,
        PATH_COUNT: 3,
        NODE_COUNT: 4,
        NODE_INCOMPATIBILITY: 5,
        CONTOUR_ORDER: 6,
        WRONG_START_POINT: 7,
        KINK: 8,
        UNDERWEIGHT: 9,
        OVERWEIGHT: 10,
        NOTHING: 11,
    }


def sort_problems(problems):
    """Sort problem groups by their most severe problem type."""
    return dict(
        sorted(
            problems.items(),
            key=lambda _: -min(
                (
                    (InterpolatableProblem.severity[p["type"]] + p.get("tolerance", 0))
                    for p in _[1]
                ),
            ),
            reverse=True,
        )
    )


def rot_list(l, k):
    """Rotate list by k items forward.  Ie. item at position 0 will be
    at position k in returned list.  Negative k is allowed."""
    return l[-k:] + l[:-k]


class PerContourPen(BasePen):
    def __init__(self, Pen, glyphset=None):
        BasePen.__init__(self, glyphset)
        self._Pen = Pen
        self._pen = None
        self.value = []

    def _moveTo(self, p0):
        self._newItem()
        self._pen.moveTo(p0)

    def _lineTo(self, p1):
        self._pen.lineTo(p1)

    def _qCurveToOne(self, p1, p2):
        self._pen.qCurveTo(p1, p2)

    def _curveToOne(self, p1, p2, p3):
        self._pen.curveTo(p1, p2, p3)

    def _closePath(self):
        self._pen.closePath()
        self._pen = None

    def _endPath(self):
        self._pen.endPath()
        self._pen = None

    def _newItem(self):
        self._pen = pen = self._Pen()
        self.value.append(pen)


class PerContourOrComponentPen(PerContourPen):
    def addComponent(self, glyphName, transformation):
        self._newItem()
        self.value[-1].addComponent(glyphName, transformation)


class SimpleRecordingPointPen(AbstractPointPen):
    def __init__(self):
        self.value = []

    def beginPath(self, identifier=None, **kwargs):
        pass

    def endPath(self) -> None:
        pass

    def addPoint(self, pt, segmentType=None):
        self.value.append((pt, False if segmentType is None else True))


def vdiff_hypot2(v0, v1):
    s = 0
    for x0, x1 in zip(v0, v1):
        d = x1 - x0
        s += d * d
    return s


def vdiff_hypot2_complex(v0, v1):
    s = 0
    for x0, x1 in zip(v0, v1):
        d = x1 - x0
        s += d.real * d.real + d.imag * d.imag
        # This does the same but seems to be slower:
        # s += (d * d.conjugate()).real
    return s


def matching_cost(G, matching):
    return sum(G[i][j] for i, j in enumerate(matching))


def min_cost_perfect_bipartite_matching_scipy(G):
    n = len(G)
    rows, cols = linear_sum_assignment(G)
    assert (rows == list(range(n))).all()
    # Convert numpy array and integer to Python types,
    # to ensure that this is JSON-serializable.
    cols = list(int(e) for e in cols)
    return list(cols), matching_cost(G, cols)


def min_cost_perfect_bipartite_matching_munkres(G):
    n = len(G)
    cols = [None] * n
    for row, col in Munkres().compute(G):
        cols[row] = col
    return cols, matching_cost(G, cols)


def min_cost_perfect_bipartite_matching_bruteforce(G):
    n = len(G)

    if n > 6:
        raise Exception("Install Python module 'munkres' or 'scipy >= 0.17.0'")

    # Otherwise just brute-force
    permutations = itertools.permutations(range(n))
    best = list(next(permutations))
    best_cost = matching_cost(G, best)
    for p in permutations:
        cost = matching_cost(G, p)
        if cost < best_cost:
            best, best_cost = list(p), cost
    return best, best_cost


# Prefer `scipy.optimize.linear_sum_assignment` for performance.
# `Munkres` is also supported as a fallback for minimalistic systems
# where installing SciPy is not feasible.
try:
    from scipy.optimize import linear_sum_assignment

    min_cost_perfect_bipartite_matching = min_cost_perfect_bipartite_matching_scipy
except ImportError:
    try:
        from munkres import Munkres

        min_cost_perfect_bipartite_matching = (
            min_cost_perfect_bipartite_matching_munkres
        )
    except ImportError:
        min_cost_perfect_bipartite_matching = (
            min_cost_perfect_bipartite_matching_bruteforce
        )


def contour_vector_from_stats(stats):
    # Don't change the order of items here.
    # It's okay to add to the end, but otherwise, other
    # code depends on it. Search for "covariance".
    size = sqrt(abs(stats.area))
    return (
        copysign((size), stats.area),
        stats.meanX,
        stats.meanY,
        stats.stddevX * 2,
        stats.stddevY * 2,
        stats.correlation * size,
    )


def matching_for_vectors(m0, m1):
    n = len(m0)

    costs = [[vdiff_hypot2(v0, v1) for v1 in m1] for v0 in m0]
    (
        matching,
        matching_cost,
    ) = min_cost_perfect_bipartite_matching(costs)
    identity_cost = sum(costs[i][i] for i in range(n))
    return matching, matching_cost, identity_cost


def points_characteristic_bits(points):
    bits = 0
    for pt, b in reversed(points):
        bits = (bits << 1) | b
    return bits


_NUM_ITEMS_PER_POINTS_COMPLEX_VECTOR = 4


def points_complex_vector(points):
    vector = []
    if not points:
        return vector
    points = [complex(*pt) for pt, _ in points]
    n = len(points)
    assert _NUM_ITEMS_PER_POINTS_COMPLEX_VECTOR == 4
    points.extend(points[: _NUM_ITEMS_PER_POINTS_COMPLEX_VECTOR - 1])
    while len(points) < _NUM_ITEMS_PER_POINTS_COMPLEX_VECTOR:
        points.extend(points[: _NUM_ITEMS_PER_POINTS_COMPLEX_VECTOR - 1])
    for i in range(n):
        # The weights are magic numbers.

        # The point itself
        p0 = points[i]
        vector.append(p0)

        # The vector to the next point
        p1 = points[i + 1]
        d0 = p1 - p0
        vector.append(d0 * 3)

        # The turn vector
        p2 = points[i + 2]
        d1 = p2 - p1
        vector.append(d1 - d0)

        # The angle to the next point, as a cross product;
        # Square root of, to match dimentionality of distance.
        cross = d0.real * d1.imag - d0.imag * d1.real
        cross = copysign(sqrt(abs(cross)), cross)
        vector.append(cross * 4)

    return vector


def add_isomorphisms(points, isomorphisms, reverse):
    reference_bits = points_characteristic_bits(points)
    n = len(points)

    # if points[0][0] == points[-1][0]:
    #   abort

    if reverse:
        points = points[::-1]
        bits = points_characteristic_bits(points)
    else:
        bits = reference_bits

    vector = points_complex_vector(points)

    assert len(vector) % n == 0
    mult = len(vector) // n
    mask = (1 << n) - 1

    for i in range(n):
        b = ((bits << (n - i)) & mask) | (bits >> i)
        if b == reference_bits:
            isomorphisms.append(
                (rot_list(vector, -i * mult), n - 1 - i if reverse else i, reverse)
            )


def find_parents_and_order(glyphsets, locations, *, discrete_axes=set()):
    parents = [None] + list(range(len(glyphsets) - 1))
    order = list(range(len(glyphsets)))
    if locations:
        # Order base master first
        bases = [
            i
            for i, l in enumerate(locations)
            if all(v == 0 for k, v in l.items() if k not in discrete_axes)
        ]
        if bases:
            log.info("Found %s base masters: %s", len(bases), bases)
        else:
            log.warning("No base master location found")

        # Form a minimum spanning tree of the locations
        try:
            from scipy.sparse.csgraph import minimum_spanning_tree

            graph = [[0] * len(locations) for _ in range(len(locations))]
            axes = set()
            for l in locations:
                axes.update(l.keys())
            axes = sorted(axes)
            vectors = [tuple(l.get(k, 0) for k in axes) for l in locations]
            for i, j in itertools.combinations(range(len(locations)), 2):
                i_discrete_location = {
                    k: v for k, v in zip(axes, vectors[i]) if k in discrete_axes
                }
                j_discrete_location = {
                    k: v for k, v in zip(axes, vectors[j]) if k in discrete_axes
                }
                if i_discrete_location != j_discrete_location:
                    continue
                graph[i][j] = vdiff_hypot2(vectors[i], vectors[j])

            tree = minimum_spanning_tree(graph, overwrite=True)
            rows, cols = tree.nonzero()
            graph = defaultdict(set)
            for row, col in zip(rows, cols):
                graph[row].add(col)
                graph[col].add(row)

            # Traverse graph from the base and assign parents
            parents = [None] * len(locations)
            order = []
            visited = set()
            queue = deque(bases)
            while queue:
                i = queue.popleft()
                visited.add(i)
                order.append(i)
                for j in sorted(graph[i]):
                    if j not in visited:
                        parents[j] = i
                        queue.append(j)
            assert len(order) == len(
                parents
            ), "Not all masters are reachable; report an issue"

        except ImportError:
            pass

        log.info("Parents: %s", parents)
        log.info("Order: %s", order)
    return parents, order



#! /usr/bin/env python3

import sys
import os
import glob
from fontTools.ttLib import identifierToTag
import textwrap


fontToolsDir = os.path.dirname(os.path.dirname(os.path.join(os.getcwd(), sys.argv[0])))
fontToolsDir = os.path.normpath(fontToolsDir)
tablesDir = os.path.join(fontToolsDir, "Lib", "fontTools", "ttLib", "tables")
docFile = os.path.join(fontToolsDir, "Doc/source/ttx.rst")

names = glob.glob1(tablesDir, "*.py")

modules = []
tables = []
for name in names:
    try:
        tag = identifierToTag(name[:-3])
    except:
        pass
    else:
        modules.append(name[:-3])
        tables.append(tag.strip())

modules.sort()
tables.sort()


with open(os.path.join(tablesDir, "__init__.py"), "w") as file:
    file.write(
        '''\
# DON'T EDIT! This file is generated by MetaTools/buildTableList.py.
def _moduleFinderHint():
    """Dummy function to let modulefinder know what tables may be
    dynamically imported. Generated by MetaTools/buildTableList.py.

        >>> _moduleFinderHint()
    """
'''
    )

    for module in modules:
        file.write("    from . import %s\n" % module)

    file.write("\n")

    file.write(
        """
if __name__ == "__main__":
    import doctest, sys

    sys.exit(doctest.testmod().failed)
"""
    )


begin = ".. begin table list\n"
end = ".. end table list"
with open(docFile) as f:
    doc = f.read()
beginPos = doc.find(begin)
assert beginPos > 0
beginPos = beginPos + len(begin) + 1
endPos = doc.find(end)

lines = textwrap.wrap(", ".join(tables[:-1]) + " and " + tables[-1], 66)
intro = "The following tables are currently supported::\n\n"
blockquote = "\n".join(" " * 4 + line for line in lines) + "\n"

doc = doc[:beginPos] + intro + blockquote + "\n" + doc[endPos:]

with open(docFile, "w") as f:
    f.write(doc)



from fontTools.ttLib import TTFont, tagToXML
from io import StringIO
import os
import sys
import re
import contextlib
import pytest

try:
    import unicodedata2
except ImportError:
    import unicodedata

    unicodedata2 = unicodedata


# Font files in data/*.{o,t}tf; output gets compared to data/*.ttx.*
TESTS = {
    "aots/base.otf": (
        "CFF ",
        "cmap",
        "head",
        "hhea",
        "hmtx",
        "maxp",
        "name",
        "OS/2",
        "post",
    ),
    "aots/classdef1_font1.otf": ("GSUB",),
    "aots/classdef1_font2.otf": ("GSUB",),
    "aots/classdef1_font3.otf": ("GSUB",),
    "aots/classdef1_font4.otf": ("GSUB",),
    "aots/classdef2_font1.otf": ("GSUB",),
    "aots/classdef2_font2.otf": ("GSUB",),
    "aots/classdef2_font3.otf": ("GSUB",),
    "aots/classdef2_font4.otf": ("GSUB",),
    "aots/cmap0_font1.otf": ("cmap",),
    "aots/cmap10_font1.otf": ("cmap",),
    "aots/cmap10_font2.otf": ("cmap",),
    "aots/cmap12_font1.otf": ("cmap",),
    "aots/cmap14_font1.otf": ("cmap",),
    "aots/cmap2_font1.otf": ("cmap",),
    "aots/cmap4_font1.otf": ("cmap",),
    "aots/cmap4_font2.otf": ("cmap",),
    "aots/cmap4_font3.otf": ("cmap",),
    "aots/cmap4_font4.otf": ("cmap",),
    "aots/cmap6_font1.otf": ("cmap",),
    "aots/cmap6_font2.otf": ("cmap",),
    "aots/cmap8_font1.otf": ("cmap",),
    "aots/cmap_composition_font1.otf": ("cmap",),
    "aots/cmap_subtableselection_font1.otf": ("cmap",),
    "aots/cmap_subtableselection_font2.otf": ("cmap",),
    "aots/cmap_subtableselection_font3.otf": ("cmap",),
    "aots/cmap_subtableselection_font4.otf": ("cmap",),
    "aots/cmap_subtableselection_font5.otf": ("cmap",),
    "aots/gpos1_1_lookupflag_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos1_1_simple_f1.otf": ("GPOS",),
    "aots/gpos1_1_simple_f2.otf": ("GPOS",),
    "aots/gpos1_1_simple_f3.otf": ("GPOS",),
    "aots/gpos1_1_simple_f4.otf": ("GPOS",),
    "aots/gpos1_2_font1.otf": ("GPOS",),
    "aots/gpos1_2_font2.otf": ("GDEF", "GPOS"),
    "aots/gpos2_1_font6.otf": ("GPOS",),
    "aots/gpos2_1_font7.otf": ("GPOS",),
    "aots/gpos2_1_lookupflag_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos2_1_lookupflag_f2.otf": ("GDEF", "GPOS"),
    "aots/gpos2_1_next_glyph_f1.otf": ("GPOS",),
    "aots/gpos2_1_next_glyph_f2.otf": ("GPOS",),
    "aots/gpos2_1_simple_f1.otf": ("GPOS",),
    "aots/gpos2_2_font1.otf": ("GPOS",),
    "aots/gpos2_2_font2.otf": ("GDEF", "GPOS"),
    "aots/gpos2_2_font3.otf": ("GDEF", "GPOS"),
    "aots/gpos2_2_font4.otf": ("GPOS",),
    "aots/gpos2_2_font5.otf": ("GPOS",),
    "aots/gpos3_font1.otf": ("GPOS",),
    "aots/gpos3_font2.otf": ("GDEF", "GPOS"),
    "aots/gpos3_font3.otf": ("GDEF", "GPOS"),
    "aots/gpos4_lookupflag_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos4_lookupflag_f2.otf": ("GDEF", "GPOS"),
    "aots/gpos4_multiple_anchors_1.otf": ("GDEF", "GPOS"),
    "aots/gpos4_simple_1.otf": ("GDEF", "GPOS"),
    "aots/gpos5_font1.otf": ("GDEF", "GPOS", "GSUB"),
    "aots/gpos6_font1.otf": ("GDEF", "GPOS"),
    "aots/gpos7_1_font1.otf": ("GPOS",),
    "aots/gpos9_font1.otf": ("GPOS",),
    "aots/gpos9_font2.otf": ("GPOS",),
    "aots/gpos_chaining1_boundary_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining1_boundary_f2.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining1_boundary_f3.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining1_boundary_f4.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining1_lookupflag_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining1_multiple_subrules_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining1_multiple_subrules_f2.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining1_next_glyph_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining1_simple_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining1_simple_f2.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining1_successive_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining2_boundary_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining2_boundary_f2.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining2_boundary_f3.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining2_boundary_f4.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining2_lookupflag_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining2_multiple_subrules_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining2_multiple_subrules_f2.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining2_next_glyph_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining2_simple_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining2_simple_f2.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining2_successive_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining3_boundary_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining3_boundary_f2.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining3_boundary_f3.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining3_boundary_f4.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining3_lookupflag_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining3_next_glyph_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining3_simple_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining3_simple_f2.otf": ("GDEF", "GPOS"),
    "aots/gpos_chaining3_successive_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_context1_boundary_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_context1_boundary_f2.otf": ("GDEF", "GPOS"),
    "aots/gpos_context1_expansion_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_context1_lookupflag_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_context1_lookupflag_f2.otf": ("GDEF", "GPOS"),
    "aots/gpos_context1_multiple_subrules_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_context1_multiple_subrules_f2.otf": ("GDEF", "GPOS"),
    "aots/gpos_context1_next_glyph_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_context1_simple_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_context1_simple_f2.otf": ("GDEF", "GPOS"),
    "aots/gpos_context1_successive_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_context2_boundary_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_context2_boundary_f2.otf": ("GDEF", "GPOS"),
    "aots/gpos_context2_classes_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_context2_classes_f2.otf": ("GDEF", "GPOS"),
    "aots/gpos_context2_expansion_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_context2_lookupflag_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_context2_lookupflag_f2.otf": ("GDEF", "GPOS"),
    "aots/gpos_context2_multiple_subrules_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_context2_multiple_subrules_f2.otf": ("GDEF", "GPOS"),
    "aots/gpos_context2_next_glyph_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_context2_simple_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_context2_simple_f2.otf": ("GDEF", "GPOS"),
    "aots/gpos_context2_successive_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_context3_boundary_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_context3_boundary_f2.otf": ("GDEF", "GPOS"),
    "aots/gpos_context3_lookupflag_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_context3_lookupflag_f2.otf": ("GDEF", "GPOS"),
    "aots/gpos_context3_next_glyph_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_context3_simple_f1.otf": ("GDEF", "GPOS"),
    "aots/gpos_context3_successive_f1.otf": ("GDEF", "GPOS"),
    "aots/gsub1_1_lookupflag_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub1_1_modulo_f1.otf": ("GSUB",),
    "aots/gsub1_1_simple_f1.otf": ("GSUB",),
    "aots/gsub1_2_lookupflag_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub1_2_simple_f1.otf": ("GSUB",),
    "aots/gsub2_1_lookupflag_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub2_1_multiple_sequences_f1.otf": ("GSUB",),
    "aots/gsub2_1_simple_f1.otf": ("GSUB",),
    "aots/gsub3_1_lookupflag_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub3_1_multiple_f1.otf": ("GSUB",),
    "aots/gsub3_1_simple_f1.otf": ("GSUB",),
    "aots/gsub4_1_lookupflag_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub4_1_multiple_ligatures_f1.otf": ("GSUB",),
    "aots/gsub4_1_multiple_ligatures_f2.otf": ("GSUB",),
    "aots/gsub4_1_multiple_ligsets_f1.otf": ("GSUB",),
    "aots/gsub4_1_simple_f1.otf": ("GSUB",),
    "aots/gsub7_font1.otf": ("GSUB",),
    "aots/gsub7_font2.otf": ("GSUB",),
    "aots/gsub_chaining1_boundary_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining1_boundary_f2.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining1_boundary_f3.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining1_boundary_f4.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining1_lookupflag_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining1_multiple_subrules_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining1_multiple_subrules_f2.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining1_next_glyph_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining1_simple_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining1_simple_f2.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining1_successive_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining2_boundary_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining2_boundary_f2.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining2_boundary_f3.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining2_boundary_f4.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining2_lookupflag_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining2_multiple_subrules_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining2_multiple_subrules_f2.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining2_next_glyph_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining2_simple_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining2_simple_f2.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining2_successive_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining3_boundary_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining3_boundary_f2.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining3_boundary_f3.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining3_boundary_f4.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining3_lookupflag_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining3_next_glyph_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining3_simple_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining3_simple_f2.otf": ("GDEF", "GSUB"),
    "aots/gsub_chaining3_successive_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_context1_boundary_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_context1_boundary_f2.otf": ("GDEF", "GSUB"),
    "aots/gsub_context1_expansion_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_context1_lookupflag_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_context1_lookupflag_f2.otf": ("GDEF", "GSUB"),
    "aots/gsub_context1_multiple_subrules_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_context1_multiple_subrules_f2.otf": ("GDEF", "GSUB"),
    "aots/gsub_context1_next_glyph_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_context1_simple_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_context1_simple_f2.otf": ("GDEF", "GSUB"),
    "aots/gsub_context1_successive_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_context2_boundary_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_context2_boundary_f2.otf": ("GDEF", "GSUB"),
    "aots/gsub_context2_classes_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_context2_classes_f2.otf": ("GDEF", "GSUB"),
    "aots/gsub_context2_expansion_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_context2_lookupflag_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_context2_lookupflag_f2.otf": ("GDEF", "GSUB"),
    "aots/gsub_context2_multiple_subrules_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_context2_multiple_subrules_f2.otf": ("GDEF", "GSUB"),
    "aots/gsub_context2_next_glyph_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_context2_simple_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_context2_simple_f2.otf": ("GDEF", "GSUB"),
    "aots/gsub_context2_successive_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_context3_boundary_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_context3_boundary_f2.otf": ("GDEF", "GSUB"),
    "aots/gsub_context3_lookupflag_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_context3_lookupflag_f2.otf": ("GDEF", "GSUB"),
    "aots/gsub_context3_next_glyph_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_context3_simple_f1.otf": ("GDEF", "GSUB"),
    "aots/gsub_context3_successive_f1.otf": ("GDEF", "GSUB"),
    "aots/lookupflag_ignore_attach_f1.otf": ("GDEF", "GSUB"),
    "aots/lookupflag_ignore_base_f1.otf": ("GDEF", "GSUB"),
    "aots/lookupflag_ignore_combination_f1.otf": ("GDEF", "GSUB"),
    "aots/lookupflag_ignore_ligatures_f1.otf": ("GDEF", "GSUB"),
    "aots/lookupflag_ignore_marks_f1.otf": ("GDEF", "GSUB"),
    "graphite/graphite_tests.ttf": ("Silf", "Glat", "Feat", "Sill"),
}


TEST_REQUIREMENTS = {
    "aots/cmap4_font4.otf": ("unicodedata2",),
}


ttLibVersion_RE = re.compile(r' ttLibVersion=".*"')


def getpath(testfile):
    path = os.path.dirname(__file__)
    return os.path.join(path, "data", testfile)


def read_expected_ttx(testfile, tableTag):
    name = os.path.splitext(testfile)[0]
    xml_expected_path = getpath("%s.ttx.%s" % (name, tagToXML(tableTag)))
    with open(xml_expected_path, "r", encoding="utf-8") as xml_file:
        xml_expected = ttLibVersion_RE.sub("", xml_file.read())
    return xml_expected


def dump_ttx(font, tableTag):
    f = StringIO()
    font.saveXML(f, tables=[tableTag])
    return ttLibVersion_RE.sub("", f.getvalue())


def load_ttx(ttx):
    f = StringIO()
    f.write(ttx)
    f.seek(0)
    font = TTFont()
    font.importXML(f)
    return font


@contextlib.contextmanager
def open_font(testfile):
    font = TTFont(getpath(testfile))
    try:
        yield font
    finally:
        font.close()


def _skip_if_requirement_missing(testfile):
    if testfile in TEST_REQUIREMENTS:
        for req in TEST_REQUIREMENTS[testfile]:
            if globals()[req] is None:
                pytest.skip("%s not installed" % req)


def test_xml_from_binary(testfile, tableTag):
    """Check XML from decompiled object."""
    _skip_if_requirement_missing(testfile)

    xml_expected = read_expected_ttx(testfile, tableTag)

    with open_font(testfile) as font:
        xml_from_binary = dump_ttx(font, tableTag)

    assert xml_expected == xml_from_binary


def test_xml_from_xml(testfile, tableTag):
    """Check XML from object read from XML."""
    _skip_if_requirement_missing(testfile)

    xml_expected = read_expected_ttx(testfile, tableTag)

    font = load_ttx(xml_expected)
    name = os.path.splitext(testfile)[0]
    setupfile = getpath("%s.ttx.%s.setup" % (name, tagToXML(tableTag)))
    if os.path.exists(setupfile):
        #        import pdb; pdb.set_trace()
        font.importXML(setupfile)
    xml_from_xml = dump_ttx(font, tableTag)

    assert xml_expected == xml_from_xml


def pytest_generate_tests(metafunc):
    # http://doc.pytest.org/en/latest/parametrize.html#basic-pytest-generate-tests-example
    fixturenames = metafunc.fixturenames
    argnames = ("testfile", "tableTag")
    if all(fn in fixturenames for fn in argnames):
        argvalues = [
            (testfile, tableTag)
            for testfile, tableTags in sorted(TESTS.items())
            for tableTag in tableTags
        ]
        metafunc.parametrize(argnames, argvalues)


if __name__ == "__main__":
    sys.exit(pytest.main(sys.argv))



from fontTools.misc.testTools import FakeFont, getXML, parseXML
from fontTools.misc.textTools import deHexStr, hexStr
from fontTools.ttLib.tables._f_v_a_r import Axis
from fontTools.ttLib import newTable, TTFont
import unittest


MVAR_DATA = deHexStr(
    "0001 0000 "  # 0:   version=1.0
    "0000 0008 "  # 4:   reserved=0, valueRecordSize=8
    "0009 "  # 8:   valueRecordCount=9
    "0054 "  # 10:  offsetToItemVariationStore=84
    "6861 7363 "  # 12:  ValueRecord.valueTag="hasc"
    "0000 "  # 16:  ValueRecord.deltaSetOuterIndex
    "0003 "  # 18:  ValueRecord.deltaSetInnerIndex
    "6863 6C61 "  # 20:  ValueRecord.valueTag="hcla"
    "0000 "  # 24:  ValueRecord.deltaSetOuterIndex
    "0003 "  # 26:  ValueRecord.deltaSetInnerIndex
    "6863 6C64 "  # 28:  ValueRecord.valueTag="hcld"
    "0000 "  # 32:  ValueRecord.deltaSetOuterIndex
    "0003 "  # 34:  ValueRecord.deltaSetInnerIndex
    "6864 7363 "  # 36:  ValueRecord.valueTag="hdsc"
    "0000 "  # 40:  ValueRecord.deltaSetOuterIndex
    "0000 "  # 42:  ValueRecord.deltaSetInnerIndex
    "686C 6770 "  # 44:  ValueRecord.valueTag="hlgp"
    "0000 "  # 48:  ValueRecord.deltaSetOuterIndex
    "0002 "  # 50:  ValueRecord.deltaSetInnerIndex
    "7362 796F "  # 52:  ValueRecord.valueTag="sbyo"
    "0000 "  # 56:  ValueRecord.deltaSetOuterIndex
    "0001 "  # 58:  ValueRecord.deltaSetInnerIndex
    "7370 796F "  # 60:  ValueRecord.valueTag="spyo"
    "0000 "  # 64:  ValueRecord.deltaSetOuterIndex
    "0002 "  # 66:  ValueRecord.deltaSetInnerIndex
    "7465 7374 "  # 68:  ValueRecord.valueTag="test"
    "0000 "  # 72:  ValueRecord.deltaSetOuterIndex
    "0002 "  # 74:  ValueRecord.deltaSetInnerIndex
    "7465 7332 "  # 76:  ValueRecord.valueTag="tes2"
    "0000 "  # 78:  ValueRecord.deltaSetOuterIndex
    "0002 "  # 82:  ValueRecord.deltaSetInnerIndex
    "0001 "  # 84:  VarStore.format=1
    "0000 000C "  # 86:  VarStore.offsetToVariationRegionList=12
    "0001 "  # 90:  VarStore.itemVariationDataCount=1
    "0000 0016 "  # 92:  VarStore.itemVariationDataOffsets[0]=22
    "0001 "  # 96:  VarRegionList.axisCount=1
    "0001 "  # 98:  VarRegionList.regionCount=1
    "0000 "  # 100:  variationRegions[0].regionAxes[0].startCoord=0.0
    "4000 "  # 102:  variationRegions[0].regionAxes[0].peakCoord=1.0
    "4000 "  # 104:  variationRegions[0].regionAxes[0].endCoord=1.0
    "0004 "  # 106:  VarData.ItemCount=4
    "0001 "  # 108:  VarData.NumShorts=1
    "0001 "  # 110:  VarData.VarRegionCount=1
    "0000 "  # 112:  VarData.VarRegionIndex[0]=0
    "FF38 "  # 114:  VarData.deltaSets[0]=-200
    "FFCE "  # 116: VarData.deltaSets[0]=-50
    "0064 "  # 118: VarData.deltaSets[0]=100
    "00C8 "  # 120: VarData.deltaSets[0]=200
)

MVAR_XML = [
    '<Version value="0x00010000"/>',
    '<Reserved value="0"/>',
    '<ValueRecordSize value="8"/>',
    "<!-- ValueRecordCount=9 -->",
    '<VarStore Format="1">',
    '  <Format value="1"/>',
    "  <VarRegionList>",
    "    <!-- RegionAxisCount=1 -->",
    "    <!-- RegionCount=1 -->",
    '    <Region index="0">',
    '      <VarRegionAxis index="0">',
    '        <StartCoord value="0.0"/>',
    '        <PeakCoord value="1.0"/>',
    '        <EndCoord value="1.0"/>',
    "      </VarRegionAxis>",
    "    </Region>",
    "  </VarRegionList>",
    "  <!-- VarDataCount=1 -->",
    '  <VarData index="0">',
    "    <!-- ItemCount=4 -->",
    '    <NumShorts value="1"/>',
    "    <!-- VarRegionCount=1 -->",
    '    <VarRegionIndex index="0" value="0"/>',
    '    <Item index="0" value="[-200]"/>',
    '    <Item index="1" value="[-50]"/>',
    '    <Item index="2" value="[100]"/>',
    '    <Item index="3" value="[200]"/>',
    "  </VarData>",
    "</VarStore>",
    '<ValueRecord index="0">',
    '  <ValueTag value="hasc"/>',
    '  <VarIdx value="3"/>',
    "</ValueRecord>",
    '<ValueRecord index="1">',
    '  <ValueTag value="hcla"/>',
    '  <VarIdx value="3"/>',
    "</ValueRecord>",
    '<ValueRecord index="2">',
    '  <ValueTag value="hcld"/>',
    '  <VarIdx value="3"/>',
    "</ValueRecord>",
    '<ValueRecord index="3">',
    '  <ValueTag value="hdsc"/>',
    '  <VarIdx value="0"/>',
    "</ValueRecord>",
    '<ValueRecord index="4">',
    '  <ValueTag value="hlgp"/>',
    '  <VarIdx value="2"/>',
    "</ValueRecord>",
    '<ValueRecord index="5">',
    '  <ValueTag value="sbyo"/>',
    '  <VarIdx value="1"/>',
    "</ValueRecord>",
    '<ValueRecord index="6">',
    '  <ValueTag value="spyo"/>',
    '  <VarIdx value="2"/>',
    "</ValueRecord>",
    '<ValueRecord index="7">',
    '  <ValueTag value="test"/>',
    '  <VarIdx value="2"/>',
    "</ValueRecord>",
    '<ValueRecord index="8">',
    '  <ValueTag value="tes2"/>',
    '  <VarIdx value="2"/>',
    "</ValueRecord>",
]


class MVARTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.maxDiff = None

    def test_decompile_toXML(self):
        mvar = newTable("MVAR")
        font = TTFont()
        mvar.decompile(MVAR_DATA, font)
        self.assertEqual(getXML(mvar.toXML), MVAR_XML)

    def test_decompile_toXML_lazy(self):
        mvar = newTable("MVAR")
        font = TTFont(lazy=True)
        mvar.decompile(MVAR_DATA, font)
        self.assertEqual(getXML(mvar.toXML), MVAR_XML)

    def test_compile_fromXML(self):
        mvar = newTable("MVAR")
        font = TTFont()
        for name, attrs, content in parseXML(MVAR_XML):
            mvar.fromXML(name, attrs, content, font=font)
        data = MVAR_DATA
        self.assertEqual(hexStr(mvar.compile(font)), hexStr(data))


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



#!/usr/bin/env python3

import argparse
import logging
import os
import sys

from fontTools.pens.cu2quPen import Cu2QuPen
from fontTools import configLogger
from fontTools.misc.cliTools import makeOutputFileName
from fontTools.pens.ttGlyphPen import TTGlyphPen
from fontTools.ttLib import TTFont, newTable


log = logging.getLogger()

# default approximation error, measured in UPEM
MAX_ERR = 1.0

# default 'post' table format
POST_FORMAT = 2.0

# assuming the input contours' direction is correctly set (counter-clockwise),
# we just flip it to clockwise
REVERSE_DIRECTION = True


def glyphs_to_quadratic(glyphs, max_err=MAX_ERR, reverse_direction=REVERSE_DIRECTION):
    quadGlyphs = {}
    for gname in glyphs.keys():
        glyph = glyphs[gname]
        ttPen = TTGlyphPen(glyphs)
        cu2quPen = Cu2QuPen(ttPen, max_err, reverse_direction=reverse_direction)
        glyph.draw(cu2quPen)
        quadGlyphs[gname] = ttPen.glyph()
    return quadGlyphs


def update_hmtx(ttFont, glyf):
    hmtx = ttFont["hmtx"]
    for glyphName, glyph in glyf.glyphs.items():
        if hasattr(glyph, "xMin"):
            hmtx[glyphName] = (hmtx[glyphName][0], glyph.xMin)


def otf_to_ttf(ttFont, post_format=POST_FORMAT, **kwargs):
    assert ttFont.sfntVersion == "OTTO"
    assert "CFF " in ttFont

    glyphOrder = ttFont.getGlyphOrder()

    ttFont["loca"] = newTable("loca")
    ttFont["glyf"] = glyf = newTable("glyf")
    glyf.glyphOrder = glyphOrder
    glyf.glyphs = glyphs_to_quadratic(ttFont.getGlyphSet(), **kwargs)
    del ttFont["CFF "]
    glyf.compile(ttFont)
    update_hmtx(ttFont, glyf)

    ttFont["maxp"] = maxp = newTable("maxp")
    maxp.tableVersion = 0x00010000
    maxp.maxZones = 1
    maxp.maxTwilightPoints = 0
    maxp.maxStorage = 0
    maxp.maxFunctionDefs = 0
    maxp.maxInstructionDefs = 0
    maxp.maxStackElements = 0
    maxp.maxSizeOfInstructions = 0
    maxp.maxComponentElements = max(
        len(g.components if hasattr(g, "components") else [])
        for g in glyf.glyphs.values()
    )
    maxp.compile(ttFont)

    post = ttFont["post"]
    post.formatType = post_format
    post.extraNames = []
    post.mapping = {}
    post.glyphOrder = glyphOrder
    try:
        post.compile(ttFont)
    except OverflowError:
        post.formatType = 3
        log.warning("Dropping glyph names, they do not fit in 'post' table.")

    ttFont.sfntVersion = "\000\001\000\000"


def main(args=None):
    configLogger(logger=log)

    parser = argparse.ArgumentParser()
    parser.add_argument("input", nargs="+", metavar="INPUT")
    parser.add_argument("-o", "--output")
    parser.add_argument("-e", "--max-error", type=float, default=MAX_ERR)
    parser.add_argument("--post-format", type=float, default=POST_FORMAT)
    parser.add_argument(
        "--keep-direction", dest="reverse_direction", action="store_false"
    )
    parser.add_argument("--face-index", type=int, default=0)
    parser.add_argument("--overwrite", action="store_true")
    options = parser.parse_args(args)

    if options.output and len(options.input) > 1:
        if not os.path.isdir(options.output):
            parser.error(
                "-o/--output option must be a directory when "
                "processing multiple fonts"
            )

    for path in options.input:
        if options.output and not os.path.isdir(options.output):
            output = options.output
        else:
            output = makeOutputFileName(
                path,
                outputDir=options.output,
                extension=".ttf",
                overWrite=options.overwrite,
            )

        font = TTFont(path, fontNumber=options.face_index)
        otf_to_ttf(
            font,
            post_format=options.post_format,
            max_err=options.max_error,
            reverse_direction=options.reverse_direction,
        )
        font.save(output)


if __name__ == "__main__":
    sys.exit(main())



import unittest

from fontTools.misc.testTools import FakeFont, getXML, parseXML
from fontTools.misc.textTools import deHexStr, hexStr
from fontTools.ttLib import newTable

POST_FORMAT_2_0_DATA = deHexStr(
    "0002 0000 "  # Version=2.0
    "0000 0000 "  # Italic Angle = 0.0
    "FF88 0028 "  # Underline position, -thickness
    "0000 0000 "  # isFixedPitch
    "0000 0000 "  # minMemType42
    "0000 0000 "  # maxMemType42
    "0000 0000 "  # minMemType1
    "0000 0000 "  # maxMemType1
    # Version 2.0 fields
    "0005 "  # numGlyphs
    "0000 0102 0103 0104 0105"  # glyphNameIndex[numGlyphs]
    "025859"  # XY
    "025859"  # XY
    "025859"  # XY
    "0458592E32"  # XY.2
)


POST_FORMAT_2_0_XML = [
    '<formatType value="2.0"/>',
    '<italicAngle value="0.0"/>',
    '<underlinePosition value="-120"/>',
    '<underlineThickness value="40"/>',
    '<isFixedPitch value="0"/>',
    '<minMemType42 value="0"/>',
    '<maxMemType42 value="0"/>',
    '<minMemType1 value="0"/>',
    '<maxMemType1 value="0"/>',
    "<psNames>",
    "  <!-- This file uses unique glyph names based on the information",
    "       found in the 'post' table. Since these names might not be unique,",
    "       we have to invent artificial names in case of clashes. In order to",
    "       be able to retain the original information, we need a name to",
    "       ps name mapping for those cases where they differ. That's what",
    "       you see below.",
    "        -->",
    '  <psName name="XY.1" psName="XY"/>',
    '  <psName name="XY.3" psName="XY"/>',
    "</psNames>",
    "<extraNames>",
    "  <!-- following are the name that are not taken from the standard Mac glyph order -->",
    '  <psName name="XY"/>',
    '  <psName name="XY"/>',
    '  <psName name="XY"/>',
    '  <psName name="XY.2"/>',
    "</extraNames>",
]


class FakeMaxp:
    def __init__(self, n: int) -> None:
        self.numGlyphs = n


class postTest(unittest.TestCase):
    def test_no_names(self):
        self.maxDiff = None
        font = FakeFont([".notdef", "XY", "XY", "XY", "XY.2"])
        font["maxp"] = FakeMaxp(5)
        table = newTable("post")
        table.decompile(POST_FORMAT_2_0_DATA, font)
        self.assertEqual(getXML(table.toXML), POST_FORMAT_2_0_XML)


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



import argparse
import logging
import sys
from io import StringIO
from pathlib import Path

from fontTools import configLogger
from fontTools.feaLib.builder import addOpenTypeFeaturesFromString
from fontTools.feaLib.error import FeatureLibError
from fontTools.feaLib.lexer import Lexer
from fontTools.misc.cliTools import makeOutputFileName
from fontTools.ttLib import TTFont, TTLibError
from fontTools.voltLib.parser import Parser
from fontTools.voltLib.voltToFea import TABLES, VoltToFea

log = logging.getLogger("fontTools.feaLib")

SUPPORTED_TABLES = TABLES + ["cmap"]


def invalid_fea_glyph_name(name):
    """Check if the glyph name is valid according to FEA syntax."""
    if name[0] not in Lexer.CHAR_NAME_START_:
        return True
    if any(c not in Lexer.CHAR_NAME_CONTINUATION_ for c in name[1:]):
        return True
    return False


def sanitize_glyph_name(name):
    """Sanitize the glyph name to ensure it is valid according to FEA syntax."""
    sanitized = ""
    for i, c in enumerate(name):
        if i == 0 and c not in Lexer.CHAR_NAME_START_:
            sanitized += "a" + c
        elif c not in Lexer.CHAR_NAME_CONTINUATION_:
            sanitized += "_"
        else:
            sanitized += c

    return sanitized


def main(args=None):
    """Build tables from a MS VOLT project into an OTF font"""
    parser = argparse.ArgumentParser(
        description="Use fontTools to compile MS VOLT projects."
    )
    parser.add_argument(
        "input",
        metavar="INPUT",
        help="Path to the input font/VTP file to process",
        type=Path,
    )
    parser.add_argument(
        "-f",
        "--font",
        metavar="INPUT_FONT",
        help="Path to the input font (if INPUT is a VTP file)",
        type=Path,
    )
    parser.add_argument(
        "-o",
        "--output",
        dest="output",
        metavar="OUTPUT",
        help="Path to the output font.",
        type=Path,
    )
    parser.add_argument(
        "-t",
        "--tables",
        metavar="TABLE_TAG",
        choices=SUPPORTED_TABLES,
        nargs="+",
        help="Specify the table(s) to be built.",
    )
    parser.add_argument(
        "-F",
        "--debug-feature-file",
        help="Write the generated feature file to disk.",
        action="store_true",
    )
    parser.add_argument(
        "--ship",
        help="Remove source VOLT tables from output font.",
        action="store_true",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="Increase the logger verbosity. Multiple -v options are allowed.",
        action="count",
        default=0,
    )
    parser.add_argument(
        "-T",
        "--traceback",
        help="show traceback for exceptions.",
        action="store_true",
    )
    options = parser.parse_args(args)

    levels = ["WARNING", "INFO", "DEBUG"]
    configLogger(level=levels[min(len(levels) - 1, options.verbose)])

    output_font = options.output or Path(
        makeOutputFileName(options.font or options.input)
    )
    log.info(f"Compiling MS VOLT to '{output_font}'")

    file_or_path = options.input
    font = None

    # If the input is a font file, extract the VOLT data from the "TSIV" table
    try:
        font = TTFont(file_or_path)
        if "TSIV" in font:
            file_or_path = StringIO(font["TSIV"].data.decode("utf-8"))
        else:
            log.error('"TSIV" table is missing')
            return 1
    except TTLibError:
        pass

    # If input is not a font file, the font must be provided
    if font is None:
        if not options.font:
            log.error("Please provide an input font")
            return 1
        font = TTFont(options.font)

    # FEA syntax does not allow some glyph names that VOLT accepts, so if we
    # found such glyph name we will temporarily rename such glyphs.
    glyphOrder = font.getGlyphOrder()
    tempGlyphOrder = None
    if any(invalid_fea_glyph_name(n) for n in glyphOrder):
        tempGlyphOrder = []
        for n in glyphOrder:
            if invalid_fea_glyph_name(n):
                n = sanitize_glyph_name(n)
                existing = set(tempGlyphOrder) | set(glyphOrder)
                while n in existing:
                    n = "a" + n
            tempGlyphOrder.append(n)
        font.setGlyphOrder(tempGlyphOrder)

    doc = Parser(file_or_path).parse()

    log.info("Converting VTP data to FEA")
    converter = VoltToFea(doc, font)
    try:
        fea = converter.convert(options.tables, ignore_unsupported_settings=True)
    except NotImplementedError as e:
        if options.traceback:
            raise
        location = getattr(e.args[0], "location", None)
        message = f'"{e}" is not supported'
        if location:
            path, line, column = location
            log.error(f"{path}:{line}:{column}: {message}")
        else:
            log.error(message)
        return 1

    fea_filename = options.input
    if options.debug_feature_file:
        fea_filename = output_font.with_suffix(".fea")
        log.info(f"Writing FEA to '{fea_filename}'")
        with open(fea_filename, "w") as fp:
            fp.write(fea)

    log.info("Compiling FEA to OpenType tables")
    try:
        addOpenTypeFeaturesFromString(
            font,
            fea,
            filename=fea_filename,
            tables=options.tables,
        )
    except FeatureLibError as e:
        if options.traceback:
            raise
        log.error(e)
        return 1

    if options.ship:
        for tag in ["TSIV", "TSIS", "TSIP", "TSID"]:
            if tag in font:
                del font[tag]

    # Restore original glyph names.
    if tempGlyphOrder:
        import io

        f = io.BytesIO()
        font.save(f)
        font = TTFont(f)
        font.setGlyphOrder(glyphOrder)
        font["post"].extraNames = []

    font.save(output_font)


if __name__ == "__main__":
    sys.exit(main())



import logging
import os
import shutil
import tempfile
import unittest
from io import open
from pathlib import Path

import pytest

from fontTools.misc.etree import XML_DECLARATION
from fontTools.pens.recordingPen import RecordingPointPen
from fontTools.ufoLib.errors import (
    GlifLibError,
    UnsupportedGLIFFormat,
    UnsupportedUFOFormat,
)
from fontTools.ufoLib.glifLib import (
    GlyphSet,
    glyphNameToFileName,
    readGlyphFromString,
    writeGlyphToString,
)

from .testSupport import getDemoFontGlyphSetPath

GLYPHSETDIR = getDemoFontGlyphSetPath()


class GlyphSetTests(unittest.TestCase):
    def setUp(self):
        self.dstDir = tempfile.mktemp()
        os.mkdir(self.dstDir)

    def tearDown(self):
        shutil.rmtree(self.dstDir)

    def testRoundTrip(self):
        import difflib

        srcDir = GLYPHSETDIR
        dstDir = self.dstDir
        src = GlyphSet(
            srcDir, ufoFormatVersion=2, validateRead=True, validateWrite=True
        )
        dst = GlyphSet(
            dstDir, ufoFormatVersion=2, validateRead=True, validateWrite=True
        )
        for glyphName in src.keys():
            g = src[glyphName]
            g.drawPoints(None)  # load attrs
            dst.writeGlyph(glyphName, g, g.drawPoints)
        # compare raw file data:
        for glyphName in sorted(src.keys()):
            fileName = src.contents[glyphName]
            with open(os.path.join(srcDir, fileName), "r") as f:
                org = f.read()
            with open(os.path.join(dstDir, fileName), "r") as f:
                new = f.read()
            added = []
            removed = []
            for line in difflib.unified_diff(org.split("\n"), new.split("\n")):
                if line.startswith("+ "):
                    added.append(line[1:])
                elif line.startswith("- "):
                    removed.append(line[1:])
            self.assertEqual(
                added, removed, "%s.glif file differs after round tripping" % glyphName
            )

    def testContentsExist(self):
        with self.assertRaises(GlifLibError):
            GlyphSet(
                self.dstDir,
                ufoFormatVersion=2,
                validateRead=True,
                validateWrite=True,
                expectContentsFile=True,
            )

    def testRebuildContents(self):
        gset = GlyphSet(GLYPHSETDIR, validateRead=True, validateWrite=True)
        contents = gset.contents
        gset.rebuildContents()
        self.assertEqual(contents, gset.contents)

    def testReverseContents(self):
        gset = GlyphSet(GLYPHSETDIR, validateRead=True, validateWrite=True)
        d = {}
        for k, v in gset.getReverseContents().items():
            d[v] = k
        org = {}
        for k, v in gset.contents.items():
            org[k] = v.lower()
        self.assertEqual(d, org)

    def testReverseContents2(self):
        src = GlyphSet(GLYPHSETDIR, validateRead=True, validateWrite=True)
        dst = GlyphSet(self.dstDir, validateRead=True, validateWrite=True)
        dstMap = dst.getReverseContents()
        self.assertEqual(dstMap, {})
        for glyphName in src.keys():
            g = src[glyphName]
            g.drawPoints(None)  # load attrs
            dst.writeGlyph(glyphName, g, g.drawPoints)
        self.assertNotEqual(dstMap, {})
        srcMap = dict(src.getReverseContents())  # copy
        self.assertEqual(dstMap, srcMap)
        del srcMap["a.glif"]
        dst.deleteGlyph("a")
        self.assertEqual(dstMap, srcMap)

    def testCustomFileNamingScheme(self):
        def myGlyphNameToFileName(glyphName, glyphSet):
            return "prefix" + glyphNameToFileName(glyphName, glyphSet)

        src = GlyphSet(GLYPHSETDIR, validateRead=True, validateWrite=True)
        dst = GlyphSet(
            self.dstDir, myGlyphNameToFileName, validateRead=True, validateWrite=True
        )
        for glyphName in src.keys():
            g = src[glyphName]
            g.drawPoints(None)  # load attrs
            dst.writeGlyph(glyphName, g, g.drawPoints)
        d = {}
        for k, v in src.contents.items():
            d[k] = "prefix" + v
        self.assertEqual(d, dst.contents)

    def testGetUnicodes(self):
        src = GlyphSet(GLYPHSETDIR, validateRead=True, validateWrite=True)
        unicodes = src.getUnicodes()
        for glyphName in src.keys():
            g = src[glyphName]
            g.drawPoints(None)  # load attrs
            if not hasattr(g, "unicodes"):
                self.assertEqual(unicodes[glyphName], [])
            else:
                self.assertEqual(g.unicodes, unicodes[glyphName])

    def testReadGlyphInvalidXml(self):
        """Test that calling readGlyph() to read a .glif with invalid XML raises
        a library error, instead of an exception from the XML dependency that is
        used internally. In addition, check that the raised exception describes
        the glyph by name and gives the location of the broken .glif file."""

        # Create a glyph set with three empty glyphs.
        glyph_set = GlyphSet(self.dstDir)
        glyph_set.writeGlyph("a", _Glyph())
        glyph_set.writeGlyph("b", _Glyph())
        glyph_set.writeGlyph("c", _Glyph())

        # Corrupt the XML of /c.
        invalid_xml = b"<abc></def>"
        Path(self.dstDir, glyph_set.contents["c"]).write_bytes(invalid_xml)

        # Confirm that reading /a and /b is fine...
        glyph_set.readGlyph("a", _Glyph())
        glyph_set.readGlyph("b", _Glyph())

        # ...but that reading /c raises a descriptive library error.
        expected_message = (
            r"GLIF contains invalid XML\.\n"
            r"The issue is in glyph 'c', located in '.*c\.glif.*\."
        )
        with pytest.raises(GlifLibError, match=expected_message):
            glyph_set.readGlyph("c", _Glyph())


class FileNameTest:
    def test_default_file_name_scheme(self):
        assert glyphNameToFileName("a", None) == "a.glif"
        assert glyphNameToFileName("A", None) == "A_.glif"
        assert glyphNameToFileName("Aring", None) == "A_ring.glif"
        assert glyphNameToFileName("F_A_B", None) == "F__A__B_.glif"
        assert glyphNameToFileName("A.alt", None) == "A_.alt.glif"
        assert glyphNameToFileName("A.Alt", None) == "A_.A_lt.glif"
        assert glyphNameToFileName(".notdef", None) == "_notdef.glif"
        assert glyphNameToFileName("T_H", None) == "T__H_.glif"
        assert glyphNameToFileName("T_h", None) == "T__h.glif"
        assert glyphNameToFileName("t_h", None) == "t_h.glif"
        assert glyphNameToFileName("F_F_I", None) == "F__F__I_.glif"
        assert glyphNameToFileName("f_f_i", None) == "f_f_i.glif"
        assert glyphNameToFileName("AE", None) == "A_E_.glif"
        assert glyphNameToFileName("Ae", None) == "A_e.glif"
        assert glyphNameToFileName("ae", None) == "ae.glif"
        assert glyphNameToFileName("aE", None) == "aE_.glif"
        assert glyphNameToFileName("a.alt", None) == "a.alt.glif"
        assert glyphNameToFileName("A.aLt", None) == "A_.aL_t.glif"
        assert glyphNameToFileName("A.alT", None) == "A_.alT_.glif"
        assert glyphNameToFileName("Aacute_V.swash", None) == "A_acute_V_.swash.glif"
        assert glyphNameToFileName(".notdef", None) == "_notdef.glif"
        assert glyphNameToFileName("con", None) == "_con.glif"
        assert glyphNameToFileName("CON", None) == "C_O_N_.glif"
        assert glyphNameToFileName("con.alt", None) == "_con.alt.glif"
        assert glyphNameToFileName("alt.con", None) == "alt._con.glif"

    def test_conflicting_case_insensitive_file_names(self, tmp_path):
        src = GlyphSet(GLYPHSETDIR)
        dst = GlyphSet(tmp_path)
        glyph = src["a"]

        dst.writeGlyph("a", glyph)
        dst.writeGlyph("A", glyph)
        dst.writeGlyph("a_", glyph)
        dst.deleteGlyph("a_")
        dst.writeGlyph("a_", glyph)
        dst.writeGlyph("A_", glyph)
        dst.writeGlyph("i_j", glyph)

        assert dst.contents == {
            "a": "a.glif",
            "A": "A_.glif",
            "a_": "a_000000000000001.glif",
            "A_": "A__.glif",
            "i_j": "i_j.glif",
        }

        # make sure filenames are unique even on case-insensitive filesystems
        assert len({fileName.lower() for fileName in dst.contents.values()}) == 5


class _Glyph:
    pass


class ReadWriteFuncTest:
    def test_roundtrip(self):
        glyph = _Glyph()
        glyph.name = "a"
        glyph.unicodes = [0x0061]

        s1 = writeGlyphToString(glyph.name, glyph)

        glyph2 = _Glyph()
        readGlyphFromString(s1, glyph2)
        assert glyph.__dict__ == glyph2.__dict__

        s2 = writeGlyphToString(glyph2.name, glyph2)
        assert s1 == s2

    def test_xml_declaration(self):
        s = writeGlyphToString("a", _Glyph())
        assert s.startswith(XML_DECLARATION % "UTF-8")

    def test_parse_xml_remove_comments(self):
        s = b"""<?xml version='1.0' encoding='UTF-8'?>
		<!-- a comment -->
		<glyph name="A" format="2">
			<advance width="1290"/>
			<unicode hex="0041"/>
			<!-- another comment -->
		</glyph>
		"""

        g = _Glyph()
        readGlyphFromString(s, g)

        assert g.name == "A"
        assert g.width == 1290
        assert g.unicodes == [0x0041]

    def test_read_invalid_xml(self):
        """Test that calling readGlyphFromString() with invalid XML raises a
        library error, instead of an exception from the XML dependency that is
        used internally."""

        invalid_xml = b"<abc></def>"
        empty_glyph = _Glyph()

        with pytest.raises(GlifLibError, match="GLIF contains invalid XML"):
            readGlyphFromString(invalid_xml, empty_glyph)

    def test_read_unsupported_format_version(self, caplog):
        s = """<?xml version='1.0' encoding='utf-8'?>
		<glyph name="A" format="0" formatMinor="0">
			<advance width="500"/>
			<unicode hex="0041"/>
		</glyph>
		"""

        with pytest.raises(UnsupportedGLIFFormat):
            readGlyphFromString(s, _Glyph())  # validate=True by default

        with pytest.raises(UnsupportedGLIFFormat):
            readGlyphFromString(s, _Glyph(), validate=True)

        caplog.clear()
        with caplog.at_level(logging.WARNING, logger="fontTools.ufoLib.glifLib"):
            readGlyphFromString(s, _Glyph(), validate=False)

        assert len(caplog.records) == 1
        assert "Unsupported GLIF format" in caplog.text
        assert "Assuming the latest supported version" in caplog.text

    def test_read_allow_format_versions(self):
        s = """<?xml version='1.0' encoding='utf-8'?>
		<glyph name="A" format="2">
			<advance width="500"/>
			<unicode hex="0041"/>
		</glyph>
		"""

        # these two calls are are equivalent
        readGlyphFromString(s, _Glyph(), formatVersions=[1, 2])
        readGlyphFromString(s, _Glyph(), formatVersions=[(1, 0), (2, 0)])

        # if at least one supported formatVersion, unsupported ones are ignored
        readGlyphFromString(s, _Glyph(), formatVersions=[(2, 0), (123, 456)])

        with pytest.raises(
            ValueError, match="None of the requested GLIF formatVersions are supported"
        ):
            readGlyphFromString(s, _Glyph(), formatVersions=[0, 2001])

        with pytest.raises(GlifLibError, match="Forbidden GLIF format version"):
            readGlyphFromString(s, _Glyph(), formatVersions=[1])

    def test_read_ensure_x_y(self):
        """Ensure that a proper GlifLibError is raised when point coordinates are
        missing, regardless of validation setting."""

        s = """<?xml version='1.0' encoding='utf-8'?>
		<glyph name="A" format="2">
			<outline>
				<contour>
					<point x="545" y="0" type="line"/>
					<point x="638" type="line"/>
				</contour>
			</outline>
		</glyph>
		"""
        pen = RecordingPointPen()

        with pytest.raises(GlifLibError, match="Required y attribute"):
            readGlyphFromString(s, _Glyph(), pen)

        with pytest.raises(GlifLibError, match="Required y attribute"):
            readGlyphFromString(s, _Glyph(), pen, validate=False)


def test_GlyphSet_unsupported_ufoFormatVersion(tmp_path, caplog):
    with pytest.raises(UnsupportedUFOFormat):
        GlyphSet(tmp_path, ufoFormatVersion=0)
    with pytest.raises(UnsupportedUFOFormat):
        GlyphSet(tmp_path, ufoFormatVersion=(0, 1))


def test_GlyphSet_writeGlyph_formatVersion(tmp_path):
    src = GlyphSet(GLYPHSETDIR)
    dst = GlyphSet(tmp_path, ufoFormatVersion=(2, 0))
    glyph = src["A"]

    # no explicit formatVersion passed: use the more recent GLIF formatVersion
    # that is supported by given ufoFormatVersion (GLIF 1 for UFO 2)
    dst.writeGlyph("A", glyph)
    glif = dst.getGLIF("A")
    assert b'format="1"' in glif
    assert b"formatMinor" not in glif  # omitted when 0

    # explicit, unknown formatVersion
    with pytest.raises(UnsupportedGLIFFormat):
        dst.writeGlyph("A", glyph, formatVersion=(0, 0))

    # explicit, known formatVersion but unsupported by given ufoFormatVersion
    with pytest.raises(
        UnsupportedGLIFFormat,
        match="Unsupported GLIF format version .*for UFO format version",
    ):
        dst.writeGlyph("A", glyph, formatVersion=(2, 0))


def test_getGLIFModificationTime():
    gs = GlyphSet(GLYPHSETDIR)
    modified = gs.getGLIFModificationTime("a")
    assert isinstance(modified, float)


def test_readGlyph_with_pointPen_and_no_glyphObject():
    """pointPen should receive outline even when glyphObject is None.

    Regression test for https://github.com/fonttools/fonttools/issues/4030
    """
    gs = GlyphSet(GLYPHSETDIR)
    pen = RecordingPointPen()
    gs.readGlyph("A", glyphObject=None, pointPen=pen)

    assert len(pen.value) > 0
    operations = [op[0] for op in pen.value]
    assert "beginPath" in operations
    assert "endPath" in operations


def test_readGlyphFromString_with_pointPen_and_no_glyphObject():
    """pointPen should receive outline even when glyphObject is None.

    Regression test for https://github.com/fonttools/fonttools/issues/4030
    """
    glif_data = b"""<?xml version='1.0' encoding='UTF-8'?>
<glyph name="test" format="2">
    <outline>
        <contour>
            <point x="0" y="0" type="line"/>
            <point x="100" y="0" type="line"/>
        </contour>
    </outline>
</glyph>
"""
    pen = RecordingPointPen()
    readGlyphFromString(glif_data, glyphObject=None, pointPen=pen)

    assert len(pen.value) > 0
    operations = [op[0] for op in pen.value]
    assert "beginPath" in operations
    assert "addPoint" in operations
    assert "endPath" in operations



from io import StringIO
from fontTools.pens.pointInsidePen import PointInsidePen
import unittest


class PointInsidePenTest(unittest.TestCase):
    def test_line(self):
        def draw_triangles(pen):
            pen.moveTo((0, 0))
            pen.lineTo((10, 5))
            pen.lineTo((10, 0))
            pen.moveTo((9, 1))
            pen.lineTo((4, 1))
            pen.lineTo((9, 4))
            pen.closePath()

        self.assertEqual(
            " *********" "   **    *" "     **  *" "       * *" "         *",
            self.render(draw_triangles, even_odd=True),
        )

        self.assertEqual(
            " *********" "   *******" "     *****" "       ***" "         *",
            self.render(draw_triangles, even_odd=False),
        )

    def test_curve(self):
        def draw_curves(pen):
            pen.moveTo((0, 0))
            pen.curveTo((9, 1), (9, 4), (0, 5))
            pen.moveTo((10, 5))
            pen.curveTo((1, 4), (1, 1), (10, 0))
            pen.closePath()

        self.assertEqual(
            "***    ***" "****  ****" "***    ***" "****  ****" "***    ***",
            self.render(draw_curves, even_odd=True),
        )

        self.assertEqual(
            "***    ***" "**********" "**********" "**********" "***    ***",
            self.render(draw_curves, even_odd=False),
        )

    def test_qCurve(self):
        def draw_qCurves(pen):
            pen.moveTo((0, 0))
            pen.qCurveTo((15, 2), (0, 5))
            pen.moveTo((10, 5))
            pen.qCurveTo((-5, 3), (10, 0))
            pen.closePath()

        self.assertEqual(
            "***     **" "****   ***" "***    ***" "***   ****" "**     ***",
            self.render(draw_qCurves, even_odd=True),
        )

        self.assertEqual(
            "***     **" "**********" "**********" "**********" "**     ***",
            self.render(draw_qCurves, even_odd=False),
        )

    @staticmethod
    def render(draw_function, even_odd):
        result = StringIO()
        for y in range(5):
            for x in range(10):
                pen = PointInsidePen(None, (x + 0.5, y + 0.5), even_odd)
                draw_function(pen)
                if pen.getResult():
                    result.write("*")
                else:
                    result.write(" ")
        return result.getvalue()

    def test_contour_no_solutions(self):
        def draw_contour(pen):
            pen.moveTo((969, 230))
            pen.curveTo((825, 348), (715, 184), (614, 202))
            pen.lineTo((614, 160))
            pen.lineTo((969, 160))
            pen.closePath()

        piPen = PointInsidePen(None, (750, 295))  # this point is outside
        draw_contour(piPen)
        self.assertEqual(piPen.getWinding(), 0)
        self.assertEqual(piPen.getResult(), False)

        piPen = PointInsidePen(None, (835, 190))  # this point is inside
        draw_contour(piPen)
        self.assertEqual(piPen.getWinding(), 1)
        self.assertEqual(piPen.getResult(), True)

    def test_contour_square_closed(self):
        def draw_contour(pen):
            pen.moveTo((100, 100))
            pen.lineTo((-100, 100))
            pen.lineTo((-100, -100))
            pen.lineTo((100, -100))
            pen.closePath()

        piPen = PointInsidePen(None, (0, 0))  # this point is inside
        draw_contour(piPen)
        self.assertEqual(piPen.getWinding(), 1)
        self.assertEqual(piPen.getResult(), True)

    def test_contour_square_opened(self):
        def draw_contour(pen):
            pen.moveTo((100, 100))
            pen.lineTo((-100, 100))
            pen.lineTo((-100, -100))
            pen.lineTo((100, -100))
            # contour not explicitly closed

        piPen = PointInsidePen(None, (0, 0))  # this point is inside
        draw_contour(piPen)
        self.assertEqual(piPen.getWinding(), 1)
        self.assertEqual(piPen.getResult(), True)

    def test_contour_circle(self):
        def draw_contour(pen):
            pen.moveTo((0, 100))
            pen.curveTo((-55, 100), (-100, 55), (-100, 0))
            pen.curveTo((-100, -55), (-55, -100), (0, -100))
            pen.curveTo((55, -100), (100, -55), (100, 0))
            pen.curveTo((100, 55), (55, 100), (0, 100))

        piPen = PointInsidePen(None, (50, 50))  # this point is inside
        draw_contour(piPen)
        self.assertEqual(piPen.getResult(), True)

        piPen = PointInsidePen(None, (50, -50))  # this point is inside
        draw_contour(piPen)
        self.assertEqual(piPen.getResult(), True)

    def test_contour_diamond(self):
        def draw_contour(pen):
            pen.moveTo((0, 100))
            pen.lineTo((100, 0))
            pen.lineTo((0, -100))
            pen.lineTo((-100, 0))
            pen.closePath()

        piPen = PointInsidePen(None, (-200, 0))  # this point is outside
        draw_contour(piPen)
        self.assertEqual(piPen.getWinding(), 0)

        piPen = PointInsidePen(None, (-200, 100))  # this point is outside
        draw_contour(piPen)
        self.assertEqual(piPen.getWinding(), 0)

        piPen = PointInsidePen(None, (-200, -100))  # this point is outside
        draw_contour(piPen)
        self.assertEqual(piPen.getWinding(), 0)

        piPen = PointInsidePen(None, (-200, 50))  # this point is outside
        draw_contour(piPen)
        self.assertEqual(piPen.getWinding(), 0)

    def test_contour_integers(self):
        def draw_contour(pen):
            pen.moveTo((728, 697))
            pen.lineTo((504, 699))
            pen.curveTo((487, 719), (508, 783), (556, 783))
            pen.lineTo((718, 783))
            pen.curveTo((739, 783), (749, 712), (728, 697))
            pen.closePath()

        piPen = PointInsidePen(None, (416, 783))  # this point is outside
        draw_contour(piPen)
        self.assertEqual(piPen.getWinding(), 0)

    def test_contour_decimals(self):
        def draw_contour(pen):
            pen.moveTo((727.546875, 697.0))
            pen.lineTo((504.375, 698.515625))
            pen.curveTo(
                (487.328125, 719.359375),
                (507.84375, 783.140625),
                (555.796875, 783.140625),
            )
            pen.lineTo((717.96875, 783.140625))
            pen.curveTo(
                (738.890625, 783.140625), (748.796875, 711.5), (727.546875, 697.0)
            )
            pen.closePath()

        piPen = PointInsidePen(None, (416.625, 783.140625))  # this point is outside
        draw_contour(piPen)
        self.assertEqual(piPen.getWinding(), 0)

    def test_contour2_integers(self):
        def draw_contour(pen):
            pen.moveTo((51, 22))
            pen.lineTo((51, 74))
            pen.lineTo((83, 50))
            pen.curveTo((83, 49), (82, 48), (82, 47))
            pen.closePath()

        piPen = PointInsidePen(None, (21, 50))  # this point is outside
        draw_contour(piPen)
        self.assertEqual(piPen.getWinding(), 0)

    def test_contour2_decimals(self):
        def draw_contour(pen):
            pen.moveTo((51.25, 21.859375))
            pen.lineTo((51.25, 73.828125))
            pen.lineTo((82.5, 50.0))
            pen.curveTo((82.5, 49.09375), (82.265625, 48.265625), (82.234375, 47.375))
            pen.closePath()

        piPen = PointInsidePen(None, (21.25, 50.0))  # this point is outside
        draw_contour(piPen)
        self.assertEqual(piPen.getWinding(), 0)


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())
